#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from scipy.stats import probplot

from evidently import ColumnMapping
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class RegErrorNormalityWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [RegressionPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = RegressionPerformanceAnalyzer.get_results(analyzers_results)

        prediction_column = results.columns.utility_columns.prediction
        target_column = results.columns.utility_columns.target

        if target_column is None or prediction_column is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )
            return None

        if self.dataset == "current":
            dataset_to_plot = (
                current_data.copy(deep=False) if current_data is not None else None
            )

        else:
            dataset_to_plot = reference_data.copy(deep=False)

        if dataset_to_plot is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires reference dataset but it is None"
                )
            return None

        dataset_to_plot.replace([np.inf, -np.inf], np.nan, inplace=True)
        dataset_to_plot.dropna(
            axis=0, how="any", inplace=True, subset=[target_column, prediction_column]
        )

        # plot error normality
        error_norm = go.Figure()

        error = dataset_to_plot[prediction_column] - dataset_to_plot[target_column]
        qq_lines = probplot(error, dist="norm", plot=None)
        theoretical_q_x = np.linspace(qq_lines[0][0][0], qq_lines[0][0][-1], 100)

        sample_quantile_trace = go.Scatter(
            x=qq_lines[0][0],
            y=qq_lines[0][1],
            mode="markers",
            name="Dataset Quantiles",
            marker=dict(size=6, color=color_options.primary_color),
        )

        theoretical_quantile_trace = go.Scatter(
            x=theoretical_q_x,
            y=qq_lines[1][0] * theoretical_q_x + qq_lines[1][1],
            mode="lines",
            name="Theoretical Quantiles",
            marker=dict(size=6, color=color_options.secondary_color),
        )

        error_norm.add_trace(sample_quantile_trace)
        error_norm.add_trace(theoretical_quantile_trace)

        error_norm.update_layout(
            xaxis_title="Theoretical Quantiles",
            yaxis_title="Dataset Quantiles",
            legend=dict(
                orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
            ),
        )

        error_norm_json = json.loads(error_norm.to_json())

        return BaseWidgetInfo(
            title=self.title,
            type="big_graph",
            size=1,
            params={
                "data": error_norm_json["data"],
                "layout": error_norm_json["layout"],
            },
        )



#!/usr/bin/env python
# coding: utf-8
from evidently.dashboard.tabs.base_tab import Tab
from evidently.dashboard.tabs.base_tab import Verbose
from evidently.dashboard.widgets.data_quality_correlations import (
    DataQualityCorrelationsWidget,
)
from evidently.dashboard.widgets.data_quality_features_widget import (
    DataQualityFeaturesWidget,
)
from evidently.dashboard.widgets.data_quality_summary_widget import (
    DataQualitySummaryWidget,
)


class DataQualityTab(Tab):
    widgets = [
        (DataQualitySummaryWidget("Data Summary"), Verbose.ALWAYS),
        (DataQualityFeaturesWidget("Features"), Verbose.ALWAYS),
        (DataQualityCorrelationsWidget("Correlations"), Verbose.FULL),
    ]



from datetime import datetime

from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceAnalyzer,
)
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceMetrics,
)
from evidently.model_profile.sections.base_profile_section import ProfileSection


class RegressionPerformanceProfileSection(ProfileSection):
    def part_id(self) -> str:
        return "regression_performance"

    def __init__(self):
        super().__init__()
        self.analyzers_types = [RegressionPerformanceAnalyzer]
        self._result = None

    def analyzers(self):
        return self.analyzers_types

    @staticmethod
    def _get_regression_performance_metrics_as_dict(
        metrics: RegressionPerformanceMetrics,
    ) -> dict:
        return {
            "mean_error": metrics.mean_error,
            "mean_abs_error": metrics.mean_abs_error,
            "mean_abs_perc_error": metrics.mean_abs_perc_error,
            "error_std": metrics.error_std,
            "abs_error_std": metrics.abs_error_std,
            "abs_perc_error_std": metrics.abs_perc_error_std,
            "error_normality": metrics.error_normality,
            "underperformance": metrics.underperformance,
        }

    def calculate(
        self, reference_data, current_data, column_mapping, analyzers_results
    ):
        result = RegressionPerformanceAnalyzer.get_results(analyzers_results)
        result_json = result.columns.as_dict()
        result_json["metrics"] = {}

        if result.error_bias is not None:
            result_json["metrics"]["error_bias"] = result.error_bias

        if result.reference_metrics is not None:
            result_json["metrics"][
                "reference"
            ] = self._get_regression_performance_metrics_as_dict(
                result.reference_metrics
            )

        if result.current_metrics is not None:
            result_json["metrics"][
                "current"
            ] = self._get_regression_performance_metrics_as_dict(result.current_metrics)

        self._result = {
            "name": self.part_id(),
            "datetime": str(datetime.now()),
            "data": result_json,
        }

    def get_results(self):
        return self._result



import abc
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Set
from typing import Type

import pandas

from evidently import ColumnMapping
from evidently.analyzers.base_analyzer import Analyzer
from evidently.options import OptionsProvider


class PipelineStage:
    _analyzers: Set[Type[Analyzer]]

    options_provider: OptionsProvider

    def __init__(self):
        self._analyzers = set()

    def add_analyzer(self, analyzer_type: Type[Analyzer]):
        self._analyzers.add(analyzer_type)

    def analyzers(self) -> Iterable[Type[Analyzer]]:
        return self._analyzers

    @abc.abstractmethod
    def calculate(
        self,
        reference_data: pandas.DataFrame,
        current_data: pandas.DataFrame,
        column_mapping: ColumnMapping,
        analyzers_results: Dict[Type[Analyzer], Any],
    ):
        raise NotImplementedError()



from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def squared_paired_dist(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Calculates the squared euclidean pairwise distance
    Args:
        x:reference data
        y:current data
    Returns:
        dist: squared euclidean pairwise distance between x and y
    """
    dotx = x**2
    doty = y**2
    dist = doty.T - 2 * np.dot(x, y.T) + dotx
    return dist


def sigma_median(dist: np.ndarray) -> float:
    """Calculate the median sigma
    Args:
        dist: pairwise distance result
    Returns:
        sigma: the median sigma
    """
    sigma = (0.5 * np.percentile(dist.flatten(), 50, interpolation="nearest")) ** 0.5
    if sigma == 0:
        return (0.5 * 1) ** 0.5
    return sigma


def rbf(x: np.ndarray, y: np.ndarray, pass_sigma: Optional[float]) -> Tuple[np.ndarray, float]:
    """compute the RBF kernel
    Args:
        x:reference data
        y:current data
        pass_sigma: override median sigma calculation (i.e provide custom sigma value)
    Returns:
        kernel: the computed kernel between reference and current
    """
    dist = squared_paired_dist(x, y)
    sigma = sigma_median(dist)
    if pass_sigma:
        sigma = pass_sigma
    gamma = 1.0 / (2.0 * sigma**2)
    return np.exp(-gamma * dist), sigma


def kernel_matrix(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Compute the kernel matrices(i.e. xx, xy and yy)
    Args:
        x: reference data
        y: current data
    Returns:
        kernel_matrix: the concatenated kernel matrix
    """
    kxx, kxxsigma = rbf(x, x, pass_sigma=None)
    kxy, tmp = rbf(x, y, pass_sigma=kxxsigma)
    kyy, tmp = rbf(y, y, pass_sigma=kxxsigma)
    kernel_matrix = np.concatenate((np.concatenate((kxx, kxy), axis=1), np.concatenate((kxy.T, kyy), axis=1)), axis=0)
    return kernel_matrix


def mmd_2samp(kernel_matrix: np.ndarray, no_y_values: int, permute: bool = False) -> float:
    """Perform the mmd test without permutation
    Args:
        kernel_matrix: the concatenated similarity matrix(i.e. xx,xy and yy)
        no_y_values: number of values in y
        permute: whether to shuffle the kernel matrix or not
    Returns:
        mmd: mmd distance without permutation
    """
    no_x_values = kernel_matrix.shape[0] - no_y_values
    if permute:
        index = np.random.permutation(kernel_matrix.shape[0])
        kernel_matrix = kernel_matrix[index][:, index]

    Kxx = kernel_matrix[:-no_y_values, :-no_y_values]
    Kyy = kernel_matrix[-no_y_values:, -no_y_values:]
    Kxy = kernel_matrix[-no_y_values:, :-no_y_values]
    A = 1 / (no_x_values * (no_x_values - 1))
    B = 1 / (no_y_values * (no_y_values - 1))
    mmd = A * Kxx.sum() + B * Kyy.sum() - 2.0 * Kxy.mean()
    return mmd


def mmd_pval(x: np.ndarray, y: np.ndarray) -> Tuple[float, float]:
    """Run the mmd test with permutations
    Args:
        x:reference data as numpy array
        y:current data as numpy array
    Returns:
        p_value:p_value
        mmd:mmd distance between x and y
    """
    kernel_mat = kernel_matrix(x, y)
    kernel_mat = kernel_mat - np.diag(np.diagonal(kernel_mat))

    mmd = mmd_2samp(kernel_mat, y.shape[0], permute=False)
    mmd_permuted = np.array([mmd_2samp(kernel_mat, y.shape[0], permute=True) for _ in range(100)])

    p_val = (mmd <= mmd_permuted).mean()

    return p_val, mmd


def _mmd_stattest(
    reference_data: pd.Series,
    current_data: pd.Series,
    feature_type: str,
    threshold: float,
) -> Tuple[float, bool]:
    """Run the  emperical maximum mean discrepancy test.
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: level of significance
    Returns:
        p_value: p-value
        test_result: whether the drift is detected
    """
    reference_data = reference_data.values.reshape(-1, 1)
    current_data = current_data.values.reshape(-1, 1)

    p_value, mmd = mmd_pval(reference_data, current_data)

    return p_value, p_value < threshold


emperical_mmd = StatTest(
    name="emperical_mmd",
    display_name="emperical_mmd",
    func=_mmd_stattest,
    allowed_feature_types=["num"],
    default_threshold=0.1,
)

register_stattest(emperical_mmd)



from typing import Generator

from evidently.analyzers.prob_classification_performance_analyzer import (
    ProbClassificationPerformanceAnalyzer,
)
from evidently.analyzers.prob_classification_performance_analyzer import (
    ProbClassificationPerformanceMetrics,
)
from evidently.model_monitoring.monitoring import MetricsType
from evidently.model_monitoring.monitoring import ModelMonitor
from evidently.model_monitoring.monitoring import ModelMonitoringMetric
from evidently.utils.data_operations import DatasetColumns


class ProbClassificationPerformanceMonitorMetricsMonitor:
    """Class for probabilistic classification performance metrics in monitor.

    Metrics list:
        - quality: model quality with macro-average metrics in `reference` and `current` datasets
        Each metric name is marked as a `metric` label
        - class_representation: quantity of items in each class
        A class name is marked as a `class_name` label
        `target` and `prediction` columns are marked as `type`
       - class_quality: quality metrics for each class
       - confusion: aggregated confusion metrics
       - class_confusion: confusion (TP, TN, FP, FN) by class
    """

    _tag = "prob_classification_performance"
    quality = ModelMonitoringMetric(f"{_tag}:quality", ["dataset", "metric"])
    class_representation = ModelMonitoringMetric(
        f"{_tag}:class_representation", ["dataset", "class_name", "type"]
    )
    class_quality = ModelMonitoringMetric(
        f"{_tag}:class_quality", ["dataset", "class_name", "metric"]
    )
    confusion = ModelMonitoringMetric(
        f"{_tag}:confusion", ["dataset", "class_x_name", "class_y_name"]
    )
    class_confusion = ModelMonitoringMetric(
        f"{_tag}:class_confusion", ["dataset", "class_name", "metric"]
    )


class ProbClassificationPerformanceMonitor(ModelMonitor):
    def monitor_id(self) -> str:
        return "prob_classification_performance"

    def analyzers(self):
        return [ProbClassificationPerformanceAnalyzer]

    @staticmethod
    def _yield_metrics(
        metrics: ProbClassificationPerformanceMetrics,
        dataset: str,
        columns: DatasetColumns,
    ) -> Generator[MetricsType, None, None]:
        yield ProbClassificationPerformanceMonitorMetricsMonitor.quality.create(
            metrics.accuracy, dict(dataset=dataset, metric="accuracy")
        )
        yield ProbClassificationPerformanceMonitorMetricsMonitor.quality.create(
            metrics.precision, dict(dataset=dataset, metric="precision")
        )
        yield ProbClassificationPerformanceMonitorMetricsMonitor.quality.create(
            metrics.recall, dict(dataset=dataset, metric="recall")
        )
        yield ProbClassificationPerformanceMonitorMetricsMonitor.quality.create(
            metrics.f1, dict(dataset=dataset, metric="f1")
        )
        yield ProbClassificationPerformanceMonitorMetricsMonitor.quality.create(
            metrics.roc_auc, dict(dataset=dataset, metric="roc_auc")
        )
        yield ProbClassificationPerformanceMonitorMetricsMonitor.quality.create(
            metrics.log_loss, dict(dataset=dataset, metric="log_loss")
        )

        # try to move classes names to readable names via ColumnMapping settings
        if columns.target_names:
            classes_names = columns.target_names

        else:
            # get classes list from the matrix data
            # remove the last 3 key - it is avg metrix values 'accuracy', 'macro avg', 'weighted avg'
            classes_names = [
                key
                for key in metrics.metrics_matrix.keys()
                if key not in ("accuracy", "macro avg", "weighted avg")
            ]

        for class_name in classes_names:
            class_name = str(class_name)
            yield ProbClassificationPerformanceMonitorMetricsMonitor.class_quality.create(
                metrics.metrics_matrix[class_name]["precision"],
                dict(dataset=dataset, class_name=class_name, metric="precision"),
            )
            yield ProbClassificationPerformanceMonitorMetricsMonitor.class_quality.create(
                metrics.metrics_matrix[class_name]["recall"],
                dict(dataset=dataset, class_name=class_name, metric="recall"),
            )
            yield ProbClassificationPerformanceMonitorMetricsMonitor.class_quality.create(
                metrics.metrics_matrix[class_name]["f1-score"],
                dict(dataset=dataset, class_name=class_name, metric="f1"),
            )

        # process confusion metrics
        for idx, class_x_name in enumerate(metrics.confusion_matrix.labels):
            class_x_name = str(class_x_name)
            yield ProbClassificationPerformanceMonitorMetricsMonitor.class_representation.create(
                sum(metrics.confusion_matrix.values[idx]),
                dict(dataset=dataset, class_name=class_x_name, type="target"),
            )
            yield ProbClassificationPerformanceMonitorMetricsMonitor.class_representation.create(
                sum([i[idx] for i in metrics.confusion_matrix.values]),
                dict(dataset=dataset, class_name=class_x_name, type="prediction"),
            )

            tp_value = metrics.confusion_by_classes[class_x_name]["tp"]
            fp_value = metrics.confusion_by_classes[class_x_name]["fp"]
            tn_value = metrics.confusion_by_classes[class_x_name]["tn"]
            fn_value = metrics.confusion_by_classes[class_x_name]["fn"]
            yield ProbClassificationPerformanceMonitorMetricsMonitor.class_confusion.create(
                tp_value, dict(dataset=dataset, class_name=class_x_name, metric="TP")
            )
            yield ProbClassificationPerformanceMonitorMetricsMonitor.class_confusion.create(
                fp_value, dict(dataset=dataset, class_name=class_x_name, metric="FP")
            )
            yield ProbClassificationPerformanceMonitorMetricsMonitor.class_confusion.create(
                tn_value, dict(dataset=dataset, class_name=class_x_name, metric="TN")
            )
            yield ProbClassificationPerformanceMonitorMetricsMonitor.class_confusion.create(
                fn_value, dict(dataset=dataset, class_name=class_x_name, metric="FN")
            )

            for idy, class_y_name in enumerate(metrics.confusion_matrix.labels):
                class_y_name = str(class_y_name)
                yield ProbClassificationPerformanceMonitorMetricsMonitor.confusion.create(
                    metrics.confusion_matrix.values[idx][idy],
                    dict(
                        dataset=dataset,
                        class_x_name=class_x_name,
                        class_y_name=class_y_name,
                    ),
                )

    def metrics(self, analyzer_results):
        results = ProbClassificationPerformanceAnalyzer.get_results(analyzer_results)

        for metric in self._yield_metrics(
            results.reference_metrics, "reference", columns=results.columns
        ):
            yield metric

        if results.current_metrics:
            for metric in self._yield_metrics(
                results.current_metrics, "current", columns=results.columns
            ):
                yield metric



"""Wasserstein distance of two samples.

Name: "wasserstein"

Import:

    >>> from evidently.calculations.stattests import wasserstein_stat_test

Properties:
- only for numerical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import wasserstein_stat_test
    >>> options = DataDriftOptions(all_features_stattest=wasserstein_stat_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="wasserstein")
"""
from typing import Tuple

import numpy as np
import pandas as pd
from scipy import stats

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def _wasserstein_distance_norm(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float
) -> Tuple[float, bool]:
    """Compute the first Wasserstein distance between two arrays normed by std of reference data
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: all values above this threshold means data drift
    Returns:
        wasserstein_distance_norm: normed Wasserstein distance
        test_result: whether the drift is detected
    """
    norm = max(np.std(reference_data), 0.001)
    wd_norm_value = stats.wasserstein_distance(reference_data, current_data) / norm
    return wd_norm_value, wd_norm_value >= threshold


wasserstein_stat_test = StatTest(
    name="wasserstein",
    display_name="Wasserstein distance (normed)",
    func=_wasserstein_distance_norm,
    allowed_feature_types=["num"],
    default_threshold=0.1,
)

register_stattest(wasserstein_stat_test)



"""Total variation distance of two samples.

Name: "TVD"

Import:

    >>> from evidently.calculations.stattests import tvd_test

Properties:
- only for numerical features
- returns distance

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import tvd_test
    >>> options = DataDriftOptions(all_features_stattest=tvd_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="TVD")
"""
from typing import Tuple

import numpy as np
import pandas as pd

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest
from evidently.calculations.stattests.utils import get_unique_not_nan_values_list_from_series
from evidently.calculations.stattests.utils import permutation_test


def _total_variation_distance(reference_data, current_data):
    """Compute the Total variation distance between two arrays
    Args:
        reference_data: reference data
        current_data: current data
    Returns:
        total_variation_distance: computed distance between reference_data and current_data
    """
    keys = get_unique_not_nan_values_list_from_series(current_data=current_data, reference_data=reference_data)
    ref_feature_dict = {**dict.fromkeys(keys, 0), **dict(reference_data.value_counts())}
    current_feature_dict = {**dict.fromkeys(keys, 0), **dict(current_data.value_counts())}
    ref = list(ref_feature_dict.values())
    curr = list(current_feature_dict.values())
    tvd = 0.5 * np.sum(np.abs(ref / sum(ref) - curr / sum(curr)))
    return tvd


def _tvd_stattest(
    reference_data: pd.Series,
    current_data: pd.Series,
    feature_type: str,
    threshold: float,
) -> Tuple[float, bool]:
    """Compute the Total variation distance (TVD) between two arrays
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: all values above this threshold means data drift
    Returns:
        p_value: two-sided p_value
        test_result: whether the drift is detected
    """

    observed = _total_variation_distance(reference_data, current_data)
    p_value = permutation_test(
        reference_data=reference_data,
        current_data=current_data,
        observed=observed,
        test_statistic_func=_total_variation_distance,
        iterations=1000,
    )

    return p_value, p_value < threshold


tvd_test = StatTest(
    name="TVD",
    display_name="Total-Variation-Distance",
    func=_tvd_stattest,
    allowed_feature_types=["cat"],
    default_threshold=0.1,
)

register_stattest(tvd_test)



import dataclasses
from typing import Dict
from typing import List
from typing import Optional

import numpy as np
import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.utils.data_operations import process_columns
from evidently.utils.visualizations import plot_line_in_time


@dataclasses.dataclass
class RegressionErrorPlotResults:
    current_scatter: Dict[str, pd.Series]
    reference_scatter: Optional[Dict[str, pd.Series]]
    x_name: str


class RegressionErrorPlot(Metric[RegressionErrorPlotResults]):
    def calculate(self, data: InputData) -> RegressionErrorPlotResults:
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction
        datetime_column_name = dataset_columns.utility_columns.date
        curr_df = data.current_data
        ref_df = data.reference_data
        if target_name is None or prediction_name is None:
            raise ValueError("The columns 'target' and 'prediction' columns should be present")
        if not isinstance(prediction_name, str):
            raise ValueError("Expect one column for prediction. List of columns was provided.")
        curr_df = self._make_df_for_plot(curr_df, target_name, prediction_name, datetime_column_name)
        current_scatter = {}
        current_scatter["Predicted - Actual"] = curr_df[prediction_name] - curr_df[target_name]
        if datetime_column_name is not None:
            current_scatter["x"] = curr_df[datetime_column_name]
            x_name = "Timestamp"
        else:
            current_scatter["x"] = curr_df.index
            x_name = "Index"

        reference_scatter: Optional[dict] = None
        if data.reference_data is not None:
            ref_df = self._make_df_for_plot(ref_df, target_name, prediction_name, datetime_column_name)
            reference_scatter = {}
            reference_scatter["Predicted - Actual"] = ref_df[prediction_name] - ref_df[target_name]
            reference_scatter["x"] = ref_df[datetime_column_name] if datetime_column_name else ref_df.index

        return RegressionErrorPlotResults(
            current_scatter=current_scatter, reference_scatter=reference_scatter, x_name=x_name
        )

    def _make_df_for_plot(self, df, target_name: str, prediction_name: str, datetime_column_name: Optional[str]):
        result = df.replace([np.inf, -np.inf], np.nan)
        if datetime_column_name is not None:
            result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name, datetime_column_name])
            return result.sort_values(datetime_column_name)
        result.dropna(axis=0, how="any", inplace=True, subset=[target_name, prediction_name])
        return result.sort_index()


@default_renderer(wrap_type=RegressionErrorPlot)
class RegressionErrorPlotRenderer(MetricRenderer):
    def render_json(self, obj: RegressionErrorPlot) -> dict:
        return {}

    def render_html(self, obj: RegressionErrorPlot) -> List[BaseWidgetInfo]:
        result = obj.get_result()
        current_scatter = result.current_scatter
        reference_scatter = None

        if result.reference_scatter is not None:
            reference_scatter = result.reference_scatter

        fig = plot_line_in_time(
            curr=current_scatter,
            ref=reference_scatter,
            x_name="x",
            y_name="Predicted - Actual",
            xaxis_name=result.x_name,
            yaxis_name="Error",
            color_options=self.color_options,
        )
        return [
            header_text(label="Error (Predicted - Actual)"),
            BaseWidgetInfo(
                title="",
                size=2,
                type="big_graph",
                params={"data": fig["data"], "layout": fig["layout"]},
            ),
        ]



#!/usr/bin/env python
# coding: utf-8

import json
import os
import uuid
from dataclasses import asdict
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import pandas

from evidently.dashboard.tabs.base_tab import Tab
from evidently.model.dashboard import DashboardInfo
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.model.widget import PlotlyGraphInfo
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.pipeline.pipeline import Pipeline
from evidently.utils import NumpyEncoder
from evidently.utils.dashboard import SaveMode
from evidently.utils.dashboard import SaveModeMap
from evidently.utils.dashboard import TemplateParams
from evidently.utils.dashboard import file_html_template
from evidently.utils.dashboard import inline_template
from evidently.utils.dashboard import save_data_file
from evidently.utils.dashboard import save_lib_files


class Dashboard(Pipeline):
    name: str
    stages: Sequence[Tab]

    def __init__(self, tabs: Sequence[Tab], options: Optional[List[object]] = None):
        super().__init__(tabs, options if options is not None else [])

    def calculate(
        self,
        reference_data: pandas.DataFrame,
        current_data: Optional[pandas.DataFrame] = None,
        column_mapping: Optional[ColumnMapping] = None,
    ):
        column_mapping = column_mapping or ColumnMapping()
        self.execute(reference_data, current_data, column_mapping)

    def __dashboard_data(self) -> Tuple[str, DashboardInfo, Dict]:
        dashboard_id = "evidently_dashboard_" + str(uuid.uuid4()).replace("-", "")
        tab_widgets = [t.info() for t in self.stages]
        dashboard_info = DashboardInfo(dashboard_id, [item for tab in tab_widgets for item in tab if item is not None])
        additional_graphs = {}
        for widget in [item for tab in tab_widgets for item in tab]:
            if widget is None:
                continue
            for graph in widget.get_additional_graphs():
                if isinstance(graph, AdditionalGraphInfo):
                    additional_graphs[graph.id] = graph.params
                elif isinstance(graph, (BaseWidgetInfo, PlotlyGraphInfo)):
                    additional_graphs[graph.id] = graph
        return dashboard_id, dashboard_info, additional_graphs

    def __render(self, dashboard_id, dashboard_info, additional_graphs, template: Callable[[TemplateParams], str]):
        return template(TemplateParams(dashboard_id, dashboard_info, additional_graphs))

    def __no_lib_render(
        self,
        dashboard_id,
        dashboard_info,
        additional_graphs,
        font_file: str,
        include_js_files: List[str],
        template: Callable[[TemplateParams], str],
    ):
        return template(
            TemplateParams(
                dashboard_id,
                dashboard_info,
                additional_graphs,
                embed_lib=False,
                embed_data=False,
                embed_font=False,
                font_file=font_file,
                include_js_files=include_js_files,
            )
        )

    def _json(self):
        dashboard_id = "evidently_dashboard_" + str(uuid.uuid4()).replace("-", "")
        tab_widgets = [t.info() for t in self.stages]
        dashboard_info = DashboardInfo(dashboard_id, [item for tab in tab_widgets for item in tab if item is not None])
        return json.dumps(asdict(dashboard_info), cls=NumpyEncoder)

    def _save_to_json(self, filename):
        parent_dir = os.path.dirname(filename)
        if parent_dir and not os.path.exists(parent_dir):
            os.makedirs(parent_dir, exist_ok=True)
        with open(filename, "w", encoding="utf-8") as out_file:
            out_file.write(self._json())

    def show(self, mode="auto"):
        dashboard_id, dashboard_info, additional_graphs = self.__dashboard_data()
        # pylint: disable=import-outside-toplevel
        render_mode = mode
        try:
            from IPython import get_ipython
            from IPython.display import HTML

            if mode == "auto":
                if type(get_ipython()).__module__.startswith("google.colab"):
                    render_mode = "inline"
                else:
                    render_mode = "nbextension"
            if render_mode == "inline":
                return HTML(self.__render(dashboard_id, dashboard_info, additional_graphs, file_html_template))
            if render_mode == "nbextension":
                return HTML(self.__render(dashboard_id, dashboard_info, additional_graphs, inline_template))
            raise ValueError(f"Unexpected value {mode}/{render_mode} for mode")
        except ImportError as err:
            raise Exception("Cannot import HTML from IPython.display, no way to show html") from err

    def html(self):
        dashboard_id, dashboard_info, additional_graphs = self.__dashboard_data()
        return self.__render(dashboard_id, dashboard_info, additional_graphs, file_html_template)

    def save(self, filename: str, mode: SaveMode = SaveMode.SINGLE_FILE):
        if isinstance(mode, str):
            _mode = SaveModeMap.get(mode)
            if _mode is None:
                raise ValueError(f"Unexpected save mode {mode}. Expected [{','.join(SaveModeMap.keys())}]")
            mode = _mode
        if mode == SaveMode.SINGLE_FILE:
            with open(filename, "w", encoding="utf-8") as out_file:
                out_file.write(self.html())
        if mode in [SaveMode.FOLDER, SaveMode.SYMLINK_FOLDER]:
            font_file, lib_file = save_lib_files(filename, mode)
            dashboard_id, dashboard_info, additional_graphs = self.__dashboard_data()
            data_file = save_data_file(filename, mode, dashboard_id, dashboard_info, additional_graphs)
            with open(filename, "w", encoding="utf-8") as out_file:
                out_file.write(
                    self.__no_lib_render(
                        dashboard_id,
                        dashboard_info,
                        additional_graphs,
                        font_file,
                        [data_file, lib_file],
                        file_html_template,
                    )
                )



"""Kolmogorov-Smirnov test of two samples.

Name: "ks"

Import:

    >>> from evidently.calculations.stattests import ks_stat_test

Properties:
- only for numerical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import ks_stat_test
    >>> options = DataDriftOptions(all_features_stattest=ks_stat_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="ks")
"""
from typing import Tuple

import pandas as pd
from scipy.stats import ks_2samp

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def _ks_stat_test(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float
) -> Tuple[float, bool]:
    """Run the two-sample Kolmogorov-Smirnov test of two samples. Alternative: two-sided
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: level of significance
    Returns:
        p_value: two-tailed p-value
        test_result: whether the drift is detected
    """
    p_value = ks_2samp(reference_data, current_data)[1]
    return p_value, p_value <= threshold


ks_stat_test = StatTest(
    name="ks", display_name="K-S p_value", func=_ks_stat_test, allowed_feature_types=["num"], default_threshold=0.05
)

register_stattest(ks_stat_test)



import dataclasses
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_quality import get_rows_count
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import table_data
from evidently.renderers.html_widgets import widget_tabs


@dataclasses.dataclass
class ValueListStat:
    number_in_list: int
    number_not_in_list: int
    share_in_list: float
    share_not_in_list: float
    values_in_list: Dict[Any, int]
    values_not_in_list: Dict[Any, int]
    rows_count: int


@dataclasses.dataclass
class ColumnValueListMetricResult:
    column_name: str
    values: List[Any]
    current: ValueListStat
    reference: Optional[ValueListStat] = None


class ColumnValueListMetric(Metric[ColumnValueListMetricResult]):
    """Calculates count and shares of values in the predefined values list"""

    column_name: str
    values: Optional[list]

    def __init__(self, column_name: str, values: Optional[list] = None) -> None:
        self.values = values
        self.column_name = column_name

    @staticmethod
    def _calculate_stats(values: list, column: pd.Series) -> ValueListStat:
        rows_count = get_rows_count(column)
        values_in_list = {}
        values_not_in_list = {}

        if rows_count == 0:
            number_in_list = 0
            number_not_in_list = 0
            share_in_list = 0.0
            share_not_in_list = 0.0

        else:
            value_counts = dict(column.value_counts(dropna=True))

            for value in value_counts:
                if value in values:
                    values_in_list[value] = value_counts[value]

                else:
                    values_not_in_list[value] = value_counts[value]

            number_in_list = sum(values_in_list.values())
            share_in_list = number_in_list / rows_count
            number_not_in_list = rows_count - number_in_list
            share_not_in_list = number_not_in_list / rows_count
            # fill other values from list with zeroes
            for value in values:
                if value not in values_in_list:
                    values_in_list[value] = 0

        return ValueListStat(
            number_in_list=number_in_list,
            number_not_in_list=number_not_in_list,
            share_in_list=share_in_list,
            share_not_in_list=share_not_in_list,
            values_in_list=values_in_list,
            values_not_in_list=values_not_in_list,
            rows_count=rows_count,
        )

    def calculate(self, data: InputData) -> ColumnValueListMetricResult:
        if data.reference_data is not None and self.column_name not in data.reference_data:
            raise ValueError(f"Column '{self.column_name}' is not in reference data.")

        if self.values is None:
            if data.reference_data is None:
                raise ValueError("Reference or values list should be present.")
            values = list(data.reference_data[self.column_name].unique())

        else:
            values = self.values

        if not values:
            raise ValueError("Values list should not be empty.")

        if self.column_name not in data.current_data:
            raise ValueError(f"Column '{self.column_name}' is not in current data.")

        current_stats = self._calculate_stats(values, data.current_data[self.column_name])

        if data.reference_data is not None:
            reference_stats: Optional[ValueListStat] = self._calculate_stats(
                values, data.reference_data[self.column_name]
            )

        else:
            reference_stats = None

        return ColumnValueListMetricResult(
            column_name=self.column_name,
            values=list(values),
            current=current_stats,
            reference=reference_stats,
        )


@default_renderer(wrap_type=ColumnValueListMetric)
class ColumnValueListMetricRenderer(MetricRenderer):
    def render_json(self, obj: ColumnValueListMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        return result

    @staticmethod
    def _get_table_stat(dataset_name: str, stats: ValueListStat) -> BaseWidgetInfo:
        matched_stat_headers = ["Value", "Count"]
        tabs = [
            TabData(
                title="In list (top-10)",
                widget=table_data(
                    title="",
                    column_names=matched_stat_headers,
                    data=[(k, v) for k, v in stats.values_in_list.items() if v > 0][:10],
                ),
            ),
            TabData(
                title="Not found (top-10)",
                widget=table_data(
                    title="",
                    column_names=matched_stat_headers,
                    data=[(k, v) for k, v in stats.values_in_list.items() if v <= 0][:10],
                ),
            ),
            TabData(
                title="Out of list (top-10)",
                widget=table_data(
                    title="",
                    column_names=matched_stat_headers,
                    data=list(stats.values_not_in_list.items())[:10],
                ),
            ),
        ]
        return widget_tabs(title=f"{dataset_name.capitalize()} dataset", tabs=tabs)

    @staticmethod
    def _get_count_info(stat: ValueListStat) -> str:
        percents = round(stat.share_in_list * 100, 3)
        return f"{stat.number_in_list} ({percents}%)"

    def _get_counters(self, metric_result: ColumnValueListMetricResult) -> BaseWidgetInfo:
        counters = [
            CounterData.string(
                label="In list (current)",
                value=self._get_count_info(metric_result.current),
            ),
        ]

        if metric_result.reference is not None:
            counters.append(
                CounterData.string(
                    label="In list (reference)",
                    value=self._get_count_info(metric_result.reference),
                ),
            )

        return counter(counters=counters)

    def render_html(self, obj: ColumnValueListMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        result = [
            header_text(label=f"Column '{metric_result.column_name}'. Value list."),
            self._get_counters(metric_result),
            self._get_table_stat("current", metric_result.current),
        ]

        if metric_result.reference:
            result.append(self._get_table_stat("reference", metric_result.reference))

        return result



"""Fisher's exact test of two samples.

Name: "fisher_exact"

Import:

    >>> from evidently.calculations.stattests import fisher_exact_test

Properties:
- only for categorical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import fisher_exact_test
    >>> options = DataDriftOptions(all_features_stattest=fisher_exact_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="fisher_exact")
"""
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.stats import fisher_exact

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest

from .utils import generate_fisher2x2_contingency_table


def _fisher_exact_stattest(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float
) -> Tuple[float, bool]:
    """Calculate the p-value of Fisher's exact test between two arrays
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: all values above this threshold means data drift
    Raises:
        ValueError: If null or inf values is found in either reference_data or current_data
        ValueError: If reference_data or current_data is not binary(unique values exceeds 2)
    Returns:
        p_value: two-tailed p-value
        test_result: whether the drift is detected
    """

    if (
        (reference_data.isnull().values.any())
        or (current_data.isnull().values.any())
        or (reference_data.isin([np.inf, -np.inf]).any())
        or (current_data.isin([np.inf, -np.inf]).any())
    ):
        raise ValueError(
            "Null or inf values found in either reference_data or current_data. Please ensure that no null or inf values are present"
        )

    if (reference_data.nunique() > 2) or (current_data.nunique() > 2):
        raise ValueError("Expects binary data for both reference and current, but found unique categories > 2")

    contingency_matrix = generate_fisher2x2_contingency_table(reference_data, current_data)
    _, p_value = fisher_exact(contingency_matrix)
    return p_value, p_value < threshold


fisher_exact_test = StatTest(
    name="fisher_exact",
    display_name="Fisher's Exact test",
    func=_fisher_exact_stattest,
    allowed_feature_types=["cat"],
    default_threshold=0.1,
)

register_stattest(fisher_exact_test)



"""Mann-Whitney U-rank test of two samples.

Name: "mannw"

Import:

    >>> from evidently.calculations.stattests import mann_whitney_u_stat_test

Properties:
- only for numerical features
- returns p-value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import mann_whitney_u_stat_test
    >>> options = DataDriftOptions(all_features_stattest=mann_whitney_u_stat_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="mannw")
"""
from typing import Tuple

import pandas as pd
from scipy.stats import mannwhitneyu

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def _mannwhitneyu_rank(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float
) -> Tuple[float, bool]:
    """Perform the Mann-Whitney U-rank test between two arrays
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: all values above this threshold means data drift
    Returns:
        pvalue: the two-tailed p-value for the test depending on alternative and method
        test_result: whether the drift is detected
    """
    p_value = mannwhitneyu(x=reference_data, y=current_data)[1]
    return p_value, p_value < threshold


mann_whitney_u_stat_test = StatTest(
    name="mannw",
    display_name="Mann-Whitney U-rank test",
    func=_mannwhitneyu_rank,
    allowed_feature_types=["num"],
    default_threshold=0.05,
)

register_stattest(mann_whitney_u_stat_test)



from evidently.base_metric import InputData
from evidently.test_preset.test_preset import TestPreset
from evidently.tests import TestValueMAE
from evidently.tests import TestValueMAPE
from evidently.tests import TestValueMeanError
from evidently.tests import TestValueRMSE
from evidently.utils.data_operations import DatasetColumns


class RegressionTestPreset(TestPreset):
    """
    Regression performance tests.

    Contains tests:
    - `TestValueMeanError`
    - `TestValueMAE`
    - `TestValueRMSE`
    - `TestValueMAPE`
    """

    def generate_tests(self, data: InputData, columns: DatasetColumns):
        return [
            TestValueMeanError(),
            TestValueMAE(),
            TestValueRMSE(),
            TestValueMAPE(),
        ]



#!/usr/bin/env python
# coding: utf-8
from evidently.dashboard.tabs.base_tab import Tab
from evidently.dashboard.tabs.base_tab import Verbose
from evidently.dashboard.widgets.class_conf_matrix_widget import ClassConfMatrixWidget
from evidently.dashboard.widgets.class_confusion_based_feature_distr_table_widget import (
    ClassConfusionBasedFeatureDistrTable,
)
from evidently.dashboard.widgets.class_metrics_matrix_widget import (
    ClassMetricsMatrixWidget,
)
from evidently.dashboard.widgets.class_quality_metrics_bar_widget import (
    ClassQualityMetricsBarWidget,
)
from evidently.dashboard.widgets.class_support_widget import ClassSupportWidget
from evidently.dashboard.widgets.target_name_widget import TargetNameWidget


class ClassificationPerformanceTab(Tab):
    widgets = [
        (
            TargetNameWidget(
                "Classification Model Performance Report.", kind="classification"
            ),
            Verbose.ALWAYS,
        ),
        (
            ClassQualityMetricsBarWidget(
                "Reference: Model Quality With Macro-average Metrics"
            ),
            Verbose.ALWAYS,
        ),
        (
            ClassQualityMetricsBarWidget(
                "Current: Model Quality With Macro-average Metrics", "current"
            ),
            Verbose.ALWAYS,
        ),
        (ClassSupportWidget("Reference: Class Representation"), Verbose.ALWAYS),
        (
            ClassSupportWidget("Current: Class Representation", "current"),
            Verbose.ALWAYS,
        ),
        (ClassConfMatrixWidget("Reference: Confusion Matrix"), Verbose.FULL),
        (ClassConfMatrixWidget("Current: Confusion Matrix", "current"), Verbose.FULL),
        (
            ClassMetricsMatrixWidget("Reference: Quality Metrics by Class"),
            Verbose.ALWAYS,
        ),
        (
            ClassMetricsMatrixWidget("Current: Quality Metrics by Class", "current"),
            Verbose.ALWAYS,
        ),
        (
            ClassConfusionBasedFeatureDistrTable("Classification Quality By Feature"),
            Verbose.FULL,
        ),
    ]



import dataclasses
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_drift import get_drift_for_columns
from evidently.calculations.stattests import PossibleStatTestType
from evidently.model.widget import BaseWidgetInfo
from evidently.options import DataDriftOptions
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import counter
from evidently.utils.data_operations import process_columns


@dataclass
class DatasetDriftMetricResults:
    drift_share: float
    number_of_columns: int
    number_of_drifted_columns: int
    share_of_drifted_columns: float
    dataset_drift: bool


class DatasetDriftMetric(Metric[DatasetDriftMetricResults]):
    columns: Optional[List[str]]
    options: DataDriftOptions
    drift_share: float

    def __init__(
        self,
        columns: Optional[List[str]] = None,
        drift_share: float = 0.5,
        stattest: Optional[PossibleStatTestType] = None,
        cat_stattest: Optional[PossibleStatTestType] = None,
        num_stattest: Optional[PossibleStatTestType] = None,
        text_stattest: Optional[PossibleStatTestType] = None,
        per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
        stattest_threshold: Optional[float] = None,
        cat_stattest_threshold: Optional[float] = None,
        num_stattest_threshold: Optional[float] = None,
        text_stattest_threshold: Optional[float] = None,
        per_column_stattest_threshold: Optional[Dict[str, float]] = None,
    ):
        self.columns = columns
        self.options = DataDriftOptions(
            all_features_stattest=stattest,
            cat_features_stattest=cat_stattest,
            num_features_stattest=num_stattest,
            text_features_stattest=text_stattest,
            per_feature_stattest=per_column_stattest,
            all_features_threshold=stattest_threshold,
            cat_features_threshold=cat_stattest_threshold,
            num_features_threshold=num_stattest_threshold,
            text_features_threshold=text_stattest_threshold,
            per_feature_threshold=per_column_stattest_threshold,
        )
        self.drift_share = drift_share

    def get_parameters(self) -> tuple:
        return self.drift_share, None if self.columns is None else tuple(self.columns), self.options

    def calculate(self, data: InputData) -> DatasetDriftMetricResults:
        if data.reference_data is None:
            raise ValueError("Reference dataset should be present")

        dataset_columns = process_columns(data.reference_data, data.column_mapping)
        result = get_drift_for_columns(
            current_data=data.current_data,
            reference_data=data.reference_data,
            data_drift_options=self.options,
            drift_share_threshold=self.drift_share,
            dataset_columns=dataset_columns,
            columns=self.columns,
        )
        return DatasetDriftMetricResults(
            drift_share=self.drift_share,
            number_of_columns=result.number_of_columns,
            number_of_drifted_columns=result.number_of_drifted_columns,
            share_of_drifted_columns=result.share_of_drifted_columns,
            dataset_drift=result.dataset_drift,
        )


@default_renderer(wrap_type=DatasetDriftMetric)
class DataDriftMetricsRenderer(MetricRenderer):
    def render_json(self, obj: DatasetDriftMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        return result

    def render_html(self, obj: DatasetDriftMetric) -> List[BaseWidgetInfo]:
        result = obj.get_result()

        if result.dataset_drift:
            drift_detected = "detected"

        else:
            drift_detected = "NOT detected"

        counters = [
            CounterData.int("Columns", result.number_of_columns),
            CounterData.int("Drifted Columns", result.number_of_drifted_columns),
            CounterData.float("Share of Drifted Columns", result.share_of_drifted_columns, 3),
        ]

        return [
            counter(
                counters=[
                    CounterData(
                        f"Dataset Drift is {drift_detected}. "
                        f"Dataset drift detection threshold is {result.drift_share}",
                        "Dataset Drift",
                    )
                ],
                title="",
            ),
            counter(
                counters=counters,
                title="",
            ),
        ]



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import pandas as pd
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from evidently import ColumnMapping
from evidently.analyzers.num_target_drift_analyzer import NumTargetDriftAnalyzer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class NumTargetPredFeatureTable(Widget):
    def analyzers(self):
        return [NumTargetDriftAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = NumTargetDriftAnalyzer.get_results(analyzers_results)

        if current_data is None:
            raise ValueError("current_data should be present")

        target_column = results.columns.utility_columns.target
        prediction_column = results.columns.utility_columns.prediction

        if target_column and prediction_column:
            return None

        additional_graphs_data = []
        params_data = []

        for feature_name in results.columns.get_all_features_list(cat_before_num=False):
            # add data for table in params
            params_data.append(
                {
                    "details": {
                        "parts": [
                            {"title": "Feature values", "id": feature_name + "_values"}
                        ],
                        "insights": [],
                    },
                    "f1": feature_name,
                }
            )

            # create plot
            fig = make_subplots(rows=1, cols=2, subplot_titles=("Reference", "Current"))

            if prediction_column is not None:
                fig.add_trace(
                    go.Scattergl(
                        x=reference_data[feature_name],
                        y=reference_data[prediction_column],
                        mode="markers",
                        name="Prediction (ref)",
                        marker=dict(size=6, color=color_options.secondary_color),
                    ),
                    row=1,
                    col=1,
                )

            if target_column is not None:
                fig.add_trace(
                    go.Scattergl(
                        x=reference_data[feature_name],
                        y=reference_data[target_column],
                        mode="markers",
                        name="Target (ref)",
                        marker=dict(size=6, color=color_options.primary_color),
                    ),
                    row=1,
                    col=1,
                )

            if prediction_column is not None:
                fig.add_trace(
                    go.Scatter(
                        x=current_data[feature_name],
                        y=current_data[prediction_column],
                        mode="markers",
                        name="Prediction (curr)",
                        marker=dict(size=6, color=color_options.secondary_color),
                    ),
                    row=1,
                    col=2,
                )

            if target_column is not None:
                fig.add_trace(
                    go.Scatter(
                        x=current_data[feature_name],
                        y=current_data[target_column],
                        mode="markers",
                        name="Target (curr)",
                        marker=dict(size=6, color=color_options.primary_color),
                    ),
                    row=1,
                    col=2,
                )

            # Update xaxis properties
            fig.update_xaxes(title_text=feature_name, showgrid=True, row=1, col=1)
            fig.update_xaxes(title_text=feature_name, showgrid=True, row=1, col=2)

            # Update yaxis properties
            fig.update_yaxes(title_text="Value", showgrid=True, row=1, col=1)
            fig.update_yaxes(title_text="Value", showgrid=True, row=1, col=2)

            fig_json = json.loads(fig.to_json())

            # write plot data in table as additional data
            additional_graphs_data.append(
                AdditionalGraphInfo(
                    feature_name + "_values",
                    {"data": fig_json["data"], "layout": fig_json["layout"]},
                )
            )

        return BaseWidgetInfo(
            title=self.title,
            type="big_table",
            size=2,
            params={
                "rowsPerPage": min(results.columns.get_features_len(), 10),
                "columns": [{"title": "Feature", "field": "f1"}],
                "data": params_data,
            },
            additionalGraphs=additional_graphs_data,
        )



#!/usr/bin/env python
# coding: utf-8

from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceAnalyzer,
)
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceMetrics,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class RegUnderperformMetricsWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.title = title
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [RegressionPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:

        results = RegressionPerformanceAnalyzer.get_results(analyzers_results)
        results_utility_columns = results.columns.utility_columns

        if (
            results_utility_columns.target is None
            or results_utility_columns.prediction is None
        ):
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )
            return None

        result_metrics = None

        if self.dataset == "current":
            result_metrics = results.current_metrics
            if result_metrics is None:
                return None

        elif self.dataset == "reference":
            result_metrics = results.reference_metrics

            if result_metrics is None:
                raise ValueError(
                    f"Widget [{self.title}] required 'reference' results from"
                    f" {RegressionPerformanceAnalyzer.__name__} but no data found"
                )

        if result_metrics is None:
            raise ValueError(
                f"Widget [{self.title}] unexpected behaviour. Var 'result_metrics should be set"
            )

        return BaseWidgetInfo(
            title=self.title,
            type="counter",
            size=2,
            params={
                "counters": [
                    {
                        "value": _format_value(result_metrics, "majority"),
                        "label": "Majority(90%)",
                    },
                    {
                        "value": _format_value(result_metrics, "underestimation"),
                        "label": "Underestimation(5%)",
                    },
                    {
                        "value": _format_value(result_metrics, "overestimation"),
                        "label": "Overestimation(5%)",
                    },
                ]
            },
        )


def _format_value(results: RegressionPerformanceMetrics, counter_type):
    return (
        f"{round(results.underperformance[counter_type]['mean_error'], 2)}"
        f" ({round(results.underperformance[counter_type]['std_error'], 2)})"
    )



#!/usr/bin/env python
# coding: utf-8
import json
from typing import Optional

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from evidently import ColumnMapping
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo


def _error_bias_string(quantile_5, quantile_95):
    def __error_bias_string(error):
        if error <= quantile_5:
            return "Underestimation"
        if error < quantile_95:
            return "Majority"
        return "Overestimation"

    return __error_bias_string


class UnderperformSegmTableWidget(Widget):
    def analyzers(self):
        return [RegressionPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:

        results = RegressionPerformanceAnalyzer.get_results(analyzers_results)
        target_name = results.columns.utility_columns.target
        prediction_name = results.columns.utility_columns.prediction

        if target_name is None or prediction_name is None:
            raise ValueError(
                f"Widget [{self.title}] requires 'target' and 'prediction' columns."
            )

        widget_info = None
        if current_data is not None:
            current_data.replace([np.inf, -np.inf], np.nan, inplace=True)
            current_data.dropna(
                axis=0, how="any", inplace=True, subset=[target_name, prediction_name]
            )

            reference_data.replace([np.inf, -np.inf], np.nan, inplace=True)
            reference_data.dropna(
                axis=0, how="any", inplace=True, subset=[target_name, prediction_name]
            )

            ref_error = reference_data[prediction_name] - reference_data[target_name]
            current_error = current_data[prediction_name] - current_data[target_name]

            ref_quntile_5 = np.quantile(ref_error, 0.05)
            ref_quntile_95 = np.quantile(ref_error, 0.95)

            current_quntile_5 = np.quantile(current_error, 0.05)
            current_quntile_95 = np.quantile(current_error, 0.95)

            # create subplots
            reference_data["dataset"] = "Reference"
            reference_data["Error bias"] = list(
                map(_error_bias_string(ref_quntile_5, ref_quntile_95), ref_error)
            )

            current_data["dataset"] = "Current"
            current_data["Error bias"] = list(
                map(
                    _error_bias_string(current_quntile_5, current_quntile_95),
                    current_error,
                )
            )
            merged_data = pd.concat([reference_data, current_data])

            reference_data.drop(["dataset", "Error bias"], axis=1, inplace=True)
            current_data.drop(["dataset", "Error bias"], axis=1, inplace=True)

            params_data = []
            additional_graphs_data = []

            for feature_name in results.columns.num_feature_names:
                feature_type = "num"

                feature_hist = px.histogram(
                    merged_data,
                    x=feature_name,
                    color="Error bias",
                    facet_col="dataset",
                    histnorm="percent",
                    barmode="overlay",
                    category_orders={
                        "dataset": ["Reference", "Current"],
                        "Error bias": ["Underestimation", "Overestimation", "Majority"],
                    },
                )

                feature_hist_json = json.loads(feature_hist.to_json())

                segment_fig = make_subplots(
                    rows=1, cols=2, subplot_titles=("Reference", "Current")
                )

                segment_fig.add_trace(
                    go.Scatter(
                        x=reference_data[target_name],
                        y=reference_data[prediction_name],
                        mode="markers",
                        marker=dict(
                            size=6,
                            cmax=max(
                                max(reference_data[feature_name]),
                                max(current_data[feature_name]),
                            ),
                            cmin=min(
                                min(reference_data[feature_name]),
                                min(current_data[feature_name]),
                            ),
                            color=reference_data[feature_name],
                        ),
                        showlegend=False,
                    ),
                    row=1,
                    col=1,
                )

                segment_fig.add_trace(
                    go.Scatter(
                        x=current_data[target_name],
                        y=current_data[prediction_name],
                        mode="markers",
                        marker=dict(
                            size=6,
                            cmax=max(
                                max(reference_data[feature_name]),
                                max(current_data[feature_name]),
                            ),
                            cmin=min(
                                min(reference_data[feature_name]),
                                min(current_data[feature_name]),
                            ),
                            color=current_data[feature_name],
                            colorbar=dict(title=feature_name),
                        ),
                        showlegend=False,
                    ),
                    row=1,
                    col=2,
                )

                # Update xaxis properties
                segment_fig.update_xaxes(
                    title_text="Actual Value", showgrid=True, row=1, col=1
                )
                segment_fig.update_xaxes(
                    title_text="Actual Value", showgrid=True, row=1, col=2
                )

                # Update yaxis properties
                segment_fig.update_yaxes(
                    title_text="Predicted Value", showgrid=True, row=1, col=1
                )
                segment_fig.update_yaxes(
                    title_text="Predicted Value", showgrid=True, row=1, col=2
                )

                segment_json = json.loads(segment_fig.to_json())

                if results.error_bias is None:
                    raise ValueError(f"Widget [{self.title}] got no error_bias value")

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "Error bias", "id": feature_name + "_hist"},
                                {
                                    "title": "Predicted vs Actual",
                                    "id": feature_name + "_segm",
                                },
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                        "f2": feature_type,
                        "f3": round(
                            results.error_bias[feature_name]["ref_majority"], 2
                        ),
                        "f4": round(results.error_bias[feature_name]["ref_under"], 2),
                        "f5": round(results.error_bias[feature_name]["ref_over"], 2),
                        "f6": round(results.error_bias[feature_name]["ref_range"], 2),
                        "f7": round(
                            results.error_bias[feature_name]["current_majority"], 2
                        ),
                        "f8": round(
                            results.error_bias[feature_name]["current_under"], 2
                        ),
                        "f9": round(
                            results.error_bias[feature_name]["current_over"], 2
                        ),
                        "f10": round(
                            results.error_bias[feature_name]["current_range"], 2
                        ),
                    }
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_hist",
                        {
                            "data": feature_hist_json["data"],
                            "layout": feature_hist_json["layout"],
                        },
                    )
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_segm",
                        {
                            "data": segment_json["data"],
                            "layout": segment_json["layout"],
                        },
                    )
                )

            for feature_name in results.columns.cat_feature_names:
                feature_type = "cat"

                feature_hist = px.histogram(
                    merged_data,
                    x=feature_name,
                    color="Error bias",
                    facet_col="dataset",
                    histnorm="percent",
                    barmode="overlay",
                    category_orders={
                        "dataset": ["Reference", "Current"],
                        "Error bias": ["Underestimation", "Overestimation", "Majority"],
                    },
                )

                feature_hist_json = json.loads(feature_hist.to_json())

                segment_fig = px.scatter(
                    merged_data,
                    x=target_name,
                    y=prediction_name,
                    color=feature_name,
                    facet_col="dataset",
                )

                segment_json = json.loads(segment_fig.to_json())

                if results.error_bias is None:
                    raise ValueError(f"Widget [{self.title}] got no error_bias value")

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "Error bias", "id": feature_name + "_hist"},
                                {
                                    "title": "Predicted vs Actual",
                                    "id": feature_name + "_segm",
                                },
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                        "f2": feature_type,
                        "f3": str(results.error_bias[feature_name]["ref_majority"]),
                        "f4": str(results.error_bias[feature_name]["ref_under"]),
                        "f5": str(results.error_bias[feature_name]["ref_over"]),
                        "f6": str(results.error_bias[feature_name]["ref_range"]),
                        "f7": str(results.error_bias[feature_name]["current_majority"]),
                        "f8": str(results.error_bias[feature_name]["current_under"]),
                        "f9": str(results.error_bias[feature_name]["current_over"]),
                        "f10": int(results.error_bias[feature_name]["current_range"]),
                    }
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_hist",
                        {
                            "data": feature_hist_json["data"],
                            "layout": feature_hist_json["layout"],
                        },
                    )
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_segm",
                        {
                            "data": segment_json["data"],
                            "layout": segment_json["layout"],
                        },
                    )
                )

            widget_info = BaseWidgetInfo(
                title=self.title,
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": min(results.columns.get_features_len(), 10),
                    "columns": [
                        {"title": "Feature", "field": "f1"},
                        {"title": "Type", "field": "f2"},
                        {"title": "REF: Majority", "field": "f3"},
                        {"title": "REF: Under", "field": "f4"},
                        {"title": "REF: Over", "field": "f5"},
                        {"title": "REF: Range(%)", "field": "f6"},
                        {"title": "CURR: Majority", "field": "f7"},
                        {"title": "CURR: Under", "field": "f8"},
                        {"title": "CURR: Over", "field": "f9"},
                        {"title": "CURR: Range(%)", "field": "f10", "sort": "desc"},
                    ],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            )

        else:
            reference_data.replace([np.inf, -np.inf], np.nan, inplace=True)
            reference_data.dropna(
                axis=0, how="any", inplace=True, subset=[target_name, prediction_name]
            )

            error = reference_data[prediction_name] - reference_data[target_name]

            quntile_5 = np.quantile(error, 0.05)
            quntile_95 = np.quantile(error, 0.95)

            reference_data["Error bias"] = list(
                map(
                    lambda x: "Underestimation"
                    if x <= quntile_5
                    else "Majority"
                    if x < quntile_95
                    else "Overestimation",
                    error,
                )
            )

            params_data = []
            additional_graphs_data = []

            for (
                feature_name
            ) in (
                results.columns.num_feature_names
            ):  # + cat_feature_names: #feature_names:

                feature_type = "num"

                hist = px.histogram(
                    reference_data,
                    x=feature_name,
                    color="Error bias",
                    histnorm="percent",
                    barmode="overlay",
                    category_orders={
                        "Error bias": ["Underestimation", "Overestimation", "Majority"]
                    },
                )

                hist_figure = json.loads(hist.to_json())

                segm = px.scatter(
                    reference_data, x=target_name, y=prediction_name, color=feature_name
                )
                segm_figure = json.loads(segm.to_json())

                if results.error_bias is None:
                    raise ValueError(f"Widget [{self.title}] got no error_bias value")

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "Error bias", "id": feature_name + "_hist"},
                                {
                                    "title": "Predicted vs Actual",
                                    "id": feature_name + "_segm",
                                },
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                        "f2": feature_type,
                        "f3": round(
                            results.error_bias[feature_name]["ref_majority"], 2
                        ),
                        "f4": round(results.error_bias[feature_name]["ref_under"], 2),
                        "f5": round(results.error_bias[feature_name]["ref_over"], 2),
                        "f6": round(results.error_bias[feature_name]["ref_range"], 2),
                    }
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_hist",
                        {"data": hist_figure["data"], "layout": hist_figure["layout"]},
                    )
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_segm",
                        {"data": segm_figure["data"], "layout": segm_figure["layout"]},
                    )
                )

            for feature_name in results.columns.cat_feature_names:  # feature_names:

                feature_type = "cat"

                hist = px.histogram(
                    reference_data,
                    x=feature_name,
                    color="Error bias",
                    histnorm="percent",
                    barmode="overlay",
                    category_orders={
                        "Error bias": ["Underestimation", "Overestimation", "Majority"]
                    },
                )

                hist_figure = json.loads(hist.to_json())

                initial_type = reference_data[feature_name].dtype
                reference_data[feature_name] = reference_data[feature_name].astype(str)
                segm = px.scatter(
                    reference_data, x=target_name, y=prediction_name, color=feature_name
                )
                reference_data[feature_name] = reference_data[feature_name].astype(
                    initial_type
                )

                segm_figure = json.loads(segm.to_json())

                if results.error_bias is None:
                    raise ValueError(f"Widget [{self.title}] got no error_bias value")

                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {"title": "Error bias", "id": feature_name + "_hist"},
                                {
                                    "title": "Predicted vs Actual",
                                    "id": feature_name + "_segm",
                                },
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                        "f2": feature_type,
                        "f3": str(results.error_bias[feature_name]["ref_majority"]),
                        "f4": str(results.error_bias[feature_name]["ref_under"]),
                        "f5": str(results.error_bias[feature_name]["ref_over"]),
                        "f6": str(results.error_bias[feature_name]["ref_range"]),
                    }
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_hist",
                        {"data": hist_figure["data"], "layout": hist_figure["layout"]},
                    )
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_segm",
                        {"data": segm_figure["data"], "layout": segm_figure["layout"]},
                    )
                )

            reference_data.drop("Error bias", axis=1, inplace=True)

            widget_info = BaseWidgetInfo(
                title=self.title,
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": min(results.columns.get_features_len(), 10),
                    "columns": [
                        {"title": "Feature", "field": "f1"},
                        {"title": "Type", "field": "f2"},
                        {"title": "Majority", "field": "f3"},
                        {"title": "Underestimation", "field": "f4"},
                        {"title": "Overestimation", "field": "f5"},
                        {"title": "Range(%)", "field": "f6", "sort": "desc"},
                    ],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            )
        return widget_info



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.analyzers.prob_classification_performance_analyzer import (
    ProbClassificationPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class ProbClassSupportWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [ProbClassificationPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = ProbClassificationPerformanceAnalyzer.get_results(analyzers_results)
        utility_columns = results.columns.utility_columns

        if utility_columns.target is None or utility_columns.prediction is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )
            return None

        if self.dataset == "reference":
            metrics = results.reference_metrics

            if metrics is None:
                raise ValueError(
                    f"Widget [{self.title}] required 'reference' results from"
                    f" {ProbClassificationPerformanceAnalyzer.__name__} but no data found"
                )

        elif self.dataset == "current":
            metrics = results.current_metrics

        else:
            raise ValueError(
                f"Widget [{self.title}] required 'current' or 'reference' dataset value"
            )

        if metrics is None:
            return None

        # plot support bar
        metrics_matrix = metrics.metrics_matrix
        metrics_frame = pd.DataFrame(metrics_matrix)

        fig = go.Figure()

        fig.add_trace(
            go.Bar(
                x=metrics_frame.columns.tolist()[:-3],
                y=metrics_frame.iloc[-1:, :-3].values[0],
                marker_color=color_options.primary_color,
                name="Support",
            )
        )

        fig.update_layout(
            xaxis_title="Class",
            yaxis_title="Number of Objects",
        )

        support_bar_json = json.loads(fig.to_json())

        return BaseWidgetInfo(
            title=self.title,
            type="big_graph",
            size=1 if current_data is not None else 2,
            params={
                "data": support_bar_json["data"],
                "layout": support_bar_json["layout"],
            },
        )



#!/usr/bin/env python
# coding: utf-8
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
import pandas as pd
from sklearn import metrics

from evidently import ColumnMapping
from evidently.analyzers.base_analyzer import Analyzer
from evidently.analyzers.base_analyzer import BaseAnalyzerResult
from evidently.calculations.classification_performance import ConfusionMatrix
from evidently.calculations.classification_performance import calculate_confusion_by_classes
from evidently.utils.data_operations import process_columns


@dataclass
class ClassificationPerformanceMetrics:
    """Class for classification performance metrics values"""

    accuracy: float
    precision: float
    recall: float
    f1: float
    metrics_matrix: Dict[str, Dict]
    confusion_matrix: ConfusionMatrix
    confusion_by_classes: Dict[Union[str, int], Dict[str, int]]


@dataclass
class ClassificationPerformanceAnalyzerResults(BaseAnalyzerResult):
    reference_metrics: Optional[ClassificationPerformanceMetrics] = None
    current_metrics: Optional[ClassificationPerformanceMetrics] = None


def classification_performance_metrics(
    target: pd.Series, prediction: pd.Series, target_names: Optional[List[str]]
) -> ClassificationPerformanceMetrics:
    # calculate metrics matrix
    metrics_matrix = metrics.classification_report(target, prediction, output_dict=True)
    # get quality metrics from the metrics matrix, do not calculate them again
    accuracy_score = metrics_matrix["accuracy"]
    avg_precision = metrics_matrix["macro avg"]["precision"]
    avg_recall = metrics_matrix["macro avg"]["recall"]
    avg_f1 = metrics_matrix["macro avg"]["f1-score"]

    # calculate confusion matrix
    confusion_matrix = metrics.confusion_matrix(target, prediction)
    # get labels from data mapping or get all values kinds from target and prediction columns
    labels = target_names if target_names else sorted(set(target) | set(prediction))
    confusion_by_classes = calculate_confusion_by_classes(confusion_matrix, labels)
    return ClassificationPerformanceMetrics(
        accuracy=accuracy_score,
        precision=avg_precision,
        recall=avg_recall,
        f1=avg_f1,
        metrics_matrix=metrics_matrix,
        confusion_matrix=ConfusionMatrix(labels=labels, values=confusion_matrix.tolist()),
        confusion_by_classes=confusion_by_classes,
    )


def _calculate_performance_metrics(
    *,
    data: pd.DataFrame,
    target_column: Union[str, Sequence[str]],
    prediction_column: Union[str, Sequence[str]],
    target_names: Optional[List[str]],
) -> ClassificationPerformanceMetrics:
    # remove all rows with infinite and NaN values from the dataset
    target_and_preds = [target_column]
    if isinstance(prediction_column, str):
        target_and_preds += [prediction_column]
    else:
        target_and_preds += prediction_column
    data.replace([np.inf, -np.inf], np.nan, inplace=True)
    data.dropna(axis=0, how="any", inplace=True, subset=target_and_preds)
    return classification_performance_metrics(data[target_column], data[prediction_column], target_names)


class ClassificationPerformanceAnalyzer(Analyzer):
    @staticmethod
    def get_results(analyzer_results) -> ClassificationPerformanceAnalyzerResults:
        return analyzer_results[ClassificationPerformanceAnalyzer]

    def calculate(
        self, reference_data: pd.DataFrame, current_data: Optional[pd.DataFrame], column_mapping: ColumnMapping
    ) -> ClassificationPerformanceAnalyzerResults:
        if reference_data is None:
            raise ValueError("reference_data should be present")

        columns = process_columns(reference_data, column_mapping)
        result = ClassificationPerformanceAnalyzerResults(columns=columns)
        target_column = columns.utility_columns.target
        prediction_column = columns.utility_columns.prediction
        target_names = columns.target_names

        if target_column is not None and prediction_column is not None:
            result.reference_metrics = _calculate_performance_metrics(
                data=reference_data,
                target_column=target_column,
                prediction_column=prediction_column,
                target_names=target_names,
            )

            if current_data is not None:
                result.current_metrics = _calculate_performance_metrics(
                    data=current_data,
                    target_column=target_column,
                    prediction_column=prediction_column,
                    target_names=target_names,
                )

        return result



"""PSI of two samples.

Name: "psi"

Import:

    >>> from evidently.calculations.stattests import psi_stat_test

Properties:
- only for categorical and numerical features
- returns PSI value

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import psi_stat_test
    >>> options = DataDriftOptions(all_features_stattest=psi_stat_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="psi")
"""
from typing import Tuple

import numpy as np
import pandas as pd

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest
from evidently.calculations.stattests.utils import get_binned_data


def _psi(
    reference_data: pd.Series, current_data: pd.Series, feature_type: str, threshold: float, n_bins: int = 30
) -> Tuple[float, bool]:
    """Calculate the PSI
    Args:
        reference_data: reference data
        current_data: current data
        feature_type: feature type
        threshold: all values above this threshold means data drift
        n_bins: number of bins
    Returns:
        psi_value: calculated PSI
        test_result: whether the drift is detected
    """
    reference_percents, current_percents = get_binned_data(reference_data, current_data, feature_type, n_bins)

    def sub_psi(ref_perc, curr_perc):
        """Calculate the actual PSI value from comparing the values.
        Update the actual value to a very small number if equal to zero
        """
        value = (ref_perc - curr_perc) * np.log(ref_perc / curr_perc)
        return value

    psi_value = 0
    for i, _ in enumerate(reference_percents):
        psi_value += sub_psi(reference_percents[i], current_percents[i])

    return psi_value, psi_value >= threshold


psi_stat_test = StatTest(
    name="psi",
    display_name="PSI",
    func=_psi,
    allowed_feature_types=["cat", "num"],
    default_threshold=0.1,
)

register_stattest(psi_stat_test)



import abc
import logging
from typing import Optional

import pandas as pd

from evidently.utils.data_preprocessing import DataDefinition


class GeneratedFeature:
    """
    Class for computation of additional features.
    """

    @abc.abstractmethod
    def generate_feature(self, data: pd.DataFrame, data_definition: DataDefinition) -> pd.DataFrame:
        """
        generate DataFrame with new features from source data.

        Returns:
            DataFrame with new features. Columns should be unique across all features of same type.
        """
        raise NotImplementedError()

    def get_parameters(self) -> Optional[tuple]:
        attributes = []
        for field, value in sorted(self.__dict__.items(), key=lambda x: x[0]):
            if isinstance(value, list):
                attributes.append(tuple(value))
            elif isinstance(value, dict):
                attributes.append(tuple([(k, value[k]) for k in sorted(value.items())]))
            else:
                attributes.append(value)
        params = tuple(attributes)
        try:
            hash(params)
        except TypeError:
            logging.warning(f"unhashable params for {type(self)}. Fallback to unique.")
            return None
        return params



import dataclasses
import json
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from plotly import graph_objs as go
from plotly.subplots import make_subplots

from evidently.options.color_scheme import ColorOptions


def plot_distr(*, hist_curr, hist_ref=None, orientation="v", color_options: ColorOptions) -> go.Figure:
    fig = go.Figure()

    fig.add_trace(
        go.Bar(
            name="current",
            x=hist_curr["x"],
            y=hist_curr["count"],
            marker_color=color_options.get_current_data_color(),
            orientation=orientation,
        )
    )
    cats = list(hist_curr["x"])
    if hist_ref is not None:
        fig.add_trace(
            go.Bar(
                name="reference",
                x=hist_ref["x"],
                y=hist_ref["count"],
                marker_color=color_options.get_reference_data_color(),
                orientation=orientation,
            )
        )
        cats = cats + list(np.setdiff1d(hist_ref["x"], cats))

    if "other" in cats:
        cats.remove("other")
        cats = cats + ["other"]
        fig.update_xaxes(categoryorder="array", categoryarray=cats)

    return fig


@dataclasses.dataclass
class Distribution:
    x: Union[np.array, list]
    y: Union[np.array, list]


def plot_distr_with_log_button(
    curr_data: pd.DataFrame,
    curr_data_log: pd.DataFrame,
    ref_data: Optional[pd.DataFrame],
    ref_data_log: Optional[pd.DataFrame],
    color_options: ColorOptions,
):
    traces = []
    visible = [True, False]
    traces.append(
        go.Bar(
            x=curr_data["x"],
            y=curr_data["count"],
            marker_color=color_options.get_current_data_color(),
            name="current",
        )
    )
    traces.append(
        go.Bar(
            x=curr_data_log["x"],
            y=curr_data_log["count"],
            visible=False,
            marker_color=color_options.get_current_data_color(),
            name="current",
        )
    )
    if ref_data is not None:
        traces.append(
            go.Bar(
                x=ref_data["x"],
                y=ref_data["count"],
                marker_color=color_options.get_reference_data_color(),
                name="reference",
            )
        )
        visible.append(True)
        if ref_data_log is not None:
            traces.append(
                go.Bar(
                    x=ref_data_log["x"],
                    y=ref_data_log["count"],
                    visible=False,
                    marker_color=color_options.get_reference_data_color(),
                    name="reference",
                )
            )
            visible.append(False)

    updatemenus = [
        dict(
            type="buttons",
            direction="right",
            x=1.0,
            yanchor="top",
            buttons=list(
                [
                    dict(
                        label="Linear Scale",
                        method="update",
                        args=[{"visible": visible}],
                    ),
                    dict(
                        label="Log Scale",
                        method="update",
                        args=[{"visible": [not x for x in visible]}],
                    ),
                ]
            ),
        )
    ]
    layout = dict(updatemenus=updatemenus)

    fig = go.Figure(data=traces, layout=layout)
    fig.update_layout(legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1))
    fig = json.loads(fig.to_json())
    return fig


def plot_distr_subplots(
    *,
    hist_curr,
    hist_ref=None,
    xaxis_name: str = "",
    yaxis_name: str = "",
    same_color: bool = False,
    color_options: ColorOptions,
):
    if same_color is None:
        curr_color = color_options.get_current_data_color()
        ref_color = color_options.get_reference_data_color()

    else:
        curr_color = color_options.get_current_data_color()
        ref_color = curr_color

    cols = 1
    subplot_titles: Union[list, str] = ""

    if hist_ref is not None:
        cols = 2
        subplot_titles = ["current", "reference"]

    fig = make_subplots(rows=1, cols=cols, shared_yaxes=True, subplot_titles=subplot_titles)
    trace = go.Bar(x=hist_curr["x"], y=hist_curr["count"], marker_color=curr_color, showlegend=False)
    fig.add_trace(trace, 1, 1)
    fig.update_xaxes(title_text=xaxis_name, row=1, col=1)

    if hist_ref is not None:
        trace = go.Bar(x=hist_ref["x"], y=hist_ref["count"], marker_color=ref_color, showlegend=False)
        fig.add_trace(trace, 1, 2)
        fig.update_xaxes(title_text=xaxis_name, row=1, col=2)
    fig.update_layout(yaxis_title=yaxis_name)
    fig = json.loads(fig.to_json())
    return fig


def plot_num_feature_in_time(
    curr_data: pd.DataFrame,
    ref_data: Optional[pd.DataFrame],
    feature_name: str,
    datetime_name: str,
    freq: str,
    color_options: ColorOptions,
):
    """
    Accepts current and reference data as pandas dataframes with two columns: datetime_name and feature_name.
    """
    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=curr_data.sort_values(datetime_name)[datetime_name],
            y=curr_data.sort_values(datetime_name)[feature_name],
            line=dict(color=color_options.get_current_data_color(), shape="spline"),
            name="current",
        )
    )
    if ref_data is not None:
        fig.add_trace(
            go.Scatter(
                x=ref_data.sort_values(datetime_name)[datetime_name],
                y=ref_data.sort_values(datetime_name)[feature_name],
                line=dict(color=color_options.get_reference_data_color(), shape="spline"),
                name="reference",
            )
        )

    fig.update_layout(yaxis_title="Mean " + feature_name + " per " + freq)
    feature_in_time_figure = json.loads(fig.to_json())
    return feature_in_time_figure


def plot_time_feature_distr(
    curr_data: pd.DataFrame, ref_data: Optional[pd.DataFrame], feature_name: str, color_options: ColorOptions
):
    """
    Accepts current and reference data as pandas dataframes with two columns: feature_name, "number_of_items"
    """
    fig = go.Figure()
    fig.add_trace(
        go.Scatter(
            x=curr_data.sort_values(feature_name)[feature_name],
            y=curr_data.sort_values(feature_name)["number_of_items"],
            line=dict(color=color_options.get_current_data_color(), shape="spline"),
            name="current",
        )
    )
    if ref_data is not None:
        fig.add_trace(
            go.Scatter(
                x=ref_data.sort_values(feature_name)[feature_name],
                y=ref_data.sort_values(feature_name)["number_of_items"],
                line=dict(color=color_options.get_reference_data_color(), shape="spline"),
                name="reference",
            )
        )
    fig.update_layout(legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1))
    fig = json.loads(fig.to_json())
    return fig


def plot_cat_feature_in_time(
    curr_data: pd.DataFrame,
    ref_data: Optional[pd.DataFrame],
    feature_name: str,
    datetime_name: str,
    freq: str,
    color_options: ColorOptions,
):
    """
    Accepts current and reference data as pandas dataframes with two columns: datetime_name and feature_name.
    """
    title = "current"
    fig = go.Figure()
    values = curr_data[feature_name].astype(str).unique()
    if ref_data is not None:
        values = np.union1d(curr_data[feature_name].astype(str).unique(), ref_data[feature_name].astype(str).unique())
    for i, val in enumerate(values):
        fig.add_trace(
            go.Bar(
                x=curr_data.loc[curr_data[feature_name].astype(str) == val, datetime_name],
                y=curr_data.loc[curr_data[feature_name].astype(str) == val, "num"],
                name=str(val),
                marker_color=color_options.color_sequence[i],
                legendgroup=str(val),
            )
        )
        if ref_data is not None:
            title = "reference/current"
            fig.add_trace(
                go.Bar(
                    x=ref_data.loc[ref_data[feature_name].astype(str) == val, datetime_name],
                    y=ref_data.loc[ref_data[feature_name].astype(str) == val, "num"],
                    name=str(val),
                    marker_color=color_options.color_sequence[i],
                    # showlegend=False,
                    legendgroup=str(val),
                    opacity=0.6,
                )
            )
    fig.update_traces(marker_line_width=0.01)
    fig.update_layout(
        barmode="stack",
        bargap=0,
        yaxis_title="count category values per " + freq,
        title=title,
    )
    feature_in_time_figure = json.loads(fig.to_json())
    return feature_in_time_figure


def plot_boxes(
    curr_for_plots: dict, ref_for_plots: Optional[dict], yaxis_title: str, xaxis_title: str, color_options: ColorOptions
):
    """
    Accepts current and reference data as dicts with box parameters ("mins", "lowers", "uppers", "means", "maxs")
    and name of boxes parameter - "values"
    """
    fig = go.Figure()
    trace = go.Box(
        lowerfence=curr_for_plots["mins"],
        q1=curr_for_plots["lowers"],
        q3=curr_for_plots["uppers"],
        median=curr_for_plots["means"],
        upperfence=curr_for_plots["maxs"],
        x=curr_for_plots["values"],
        name="current",
        marker_color=color_options.get_current_data_color(),
    )
    fig.add_trace(trace)
    if ref_for_plots is not None:
        trace = go.Box(
            lowerfence=curr_for_plots["mins"],
            q1=ref_for_plots["lowers"],
            q3=ref_for_plots["uppers"],
            median=ref_for_plots["means"],
            upperfence=ref_for_plots["maxs"],
            x=ref_for_plots["values"],
            name="reference",
            marker_color=color_options.get_reference_data_color(),
        )
        fig.add_trace(trace)
        fig.update_layout(boxmode="group")
    fig.update_layout(yaxis_title=yaxis_title, xaxis_title=xaxis_title, boxmode="group")
    fig = json.loads(fig.to_json())
    return fig


def make_hist_for_num_plot(curr: pd.Series, ref: pd.Series = None):
    result = {}
    if ref is not None:
        ref = ref.dropna()
    bins = np.histogram_bin_edges(pd.concat([curr.dropna(), ref]), bins="doane")
    curr_hist = np.histogram(curr, bins=bins)
    result["current"] = make_hist_df(curr_hist)
    if ref is not None:
        ref_hist = np.histogram(ref, bins=bins)
        result["reference"] = make_hist_df(ref_hist)
    return result


def plot_cat_cat_rel(
    curr: pd.DataFrame, ref: pd.DataFrame, target_name: str, feature_name: str, color_options: ColorOptions
):
    """
    Accepts current and reference data as pandas dataframes with two columns: feature_name and "count_objects".
    """
    cols = 1
    subplot_titles: Union[list, str] = ""
    if ref is not None:
        cols = 2
        subplot_titles = ["current", "reference"]
    fig = make_subplots(rows=1, cols=cols, shared_yaxes=True, subplot_titles=subplot_titles)
    for i, val in enumerate(curr[target_name].astype(str).unique()):
        trace = go.Bar(
            x=curr.loc[curr[target_name] == val, feature_name],
            y=curr.loc[curr[target_name] == val, "count_objects"],
            marker_color=color_options.color_sequence[i],
            name=str(val),
            legendgroup=str(val),
        )
        fig.add_trace(trace, 1, 1)

    if ref is not None:
        for i, val in enumerate(ref[target_name].astype(str).unique()):
            trace = go.Bar(
                x=ref.loc[ref[target_name] == val, feature_name],
                y=ref.loc[ref[target_name] == val, "count_objects"],
                marker_color=color_options.color_sequence[i],
                opacity=0.6,
                name=str(val),
                legendgroup=str(val),
            )
            fig.add_trace(trace, 1, 2)
    fig.update_layout(yaxis_title="count")
    fig = json.loads(fig.to_json())
    return fig


def plot_num_num_rel(
    curr: Dict[str, list],
    ref: Optional[Dict[str, list]],
    target_name: str,
    column_name: str,
    color_options: ColorOptions,
):
    cols = 1
    if ref is not None:
        cols = 2
    fig = make_subplots(rows=1, cols=cols, shared_yaxes=True)
    trace = go.Scatter(
        x=curr[column_name],
        y=curr[target_name],
        mode="markers",
        marker_color=color_options.get_current_data_color(),
        name="current",
    )
    fig.add_trace(trace, 1, 1)
    fig.update_xaxes(title_text=column_name, row=1, col=1)
    if ref is not None:
        trace = go.Scatter(
            x=ref[column_name],
            y=ref[target_name],
            mode="markers",
            marker_color=color_options.get_reference_data_color(),
            name="reference",
        )
        fig.add_trace(trace, 1, 2)
        fig.update_xaxes(title_text=column_name, row=1, col=2)
    fig.update_layout(yaxis_title=target_name, legend={"itemsizing": "constant"})
    fig.update_traces(marker_size=4)
    fig = json.loads(fig.to_json())
    return fig


def make_hist_for_cat_plot(
    curr: pd.Series, ref: pd.Series = None, normalize: bool = False, dropna=False
) -> Dict[str, pd.Series]:
    result = {}
    hist_df = curr.astype(str).value_counts(normalize=normalize, dropna=dropna).reset_index()
    hist_df.columns = ["x", "count"]
    result["current"] = hist_df
    if ref is not None:
        hist_df = ref.astype(str).value_counts(normalize=normalize, dropna=dropna).reset_index()
        hist_df.columns = ["x", "count"]
        result["reference"] = hist_df
    return result


def get_distribution_for_category_column(column: pd.Series, normalize: bool = False) -> Distribution:
    value_counts = column.value_counts(normalize=normalize, dropna=False)
    return Distribution(
        x=value_counts.index.values,
        y=value_counts.values,
    )


def get_distribution_for_numerical_column(
    column: pd.Series,
    bins: Optional[Union[list, np.array]] = None,
) -> Distribution:
    if bins is None:
        bins = np.histogram_bin_edges(column, bins="doane")

    histogram = np.histogram(column, bins=bins)
    return Distribution(
        x=histogram[1],
        y=histogram[0],
    )


def get_distribution_for_column(
    *, column_type: str, current: pd.Series, reference: Optional[pd.Series] = None
) -> Tuple[Distribution, Optional[Distribution]]:
    reference_distribution: Optional[Distribution] = None

    if column_type == "cat":
        current_distribution = get_distribution_for_category_column(current)

        if reference is not None:
            reference_distribution = get_distribution_for_category_column(reference)

    elif column_type == "num":
        if reference is not None:
            bins = np.histogram_bin_edges(pd.concat([current.dropna(), reference.dropna()]), bins="doane")
            reference_distribution = get_distribution_for_numerical_column(reference, bins)

        else:
            bins = np.histogram_bin_edges(current.dropna(), bins="doane")

        current_distribution = get_distribution_for_numerical_column(current, bins)

    else:
        raise ValueError(f"Cannot get distribution for a column with type {column_type}")

    return current_distribution, reference_distribution


def make_hist_df(hist: Tuple[np.array, np.array]) -> pd.DataFrame:
    hist_df = pd.DataFrame(
        np.array([hist[1][:-1], hist[0], [f"{x[0]}-{x[1]}" for x in zip(hist[1][:-1], hist[1][1:])]]).T,
        columns=["x", "count", "range"],
    )

    hist_df["x"] = hist_df["x"].astype(float)
    hist_df["count"] = hist_df["count"].astype(int)
    return hist_df


def plot_scatter(
    *,
    curr: Dict[str, Union[list, pd.Series]],
    ref: Optional[Dict[str, list]],
    x: str,
    y: str,
    xaxis_name: str = None,
    yaxis_name: str = None,
    color_options: ColorOptions,
):
    cols = 1
    if xaxis_name is None:
        xaxis_name = x
    if yaxis_name is None:
        yaxis_name = y
    if ref is not None:
        cols = 2
    fig = make_subplots(rows=1, cols=cols, shared_yaxes=True)
    trace = go.Scatter(
        x=curr[x],
        y=curr[y],
        mode="markers",
        marker_color=color_options.get_current_data_color(),
        name="current",
    )
    fig.add_trace(trace, 1, 1)
    fig.update_xaxes(title_text=xaxis_name, row=1, col=1)
    if ref is not None:
        trace = go.Scatter(
            x=ref[x],
            y=ref[y],
            mode="markers",
            marker_color=color_options.get_reference_data_color(),
            name="reference",
        )
        fig.add_trace(trace, 1, 2)
        fig.update_xaxes(title_text=xaxis_name, row=1, col=2)
    fig.update_layout(yaxis_title=yaxis_name, legend={"itemsizing": "constant"})
    fig.update_traces(marker_size=4)
    fig = json.loads(fig.to_json())
    return fig


def plot_pred_actual_time(
    *,
    curr: Dict[str, pd.Series],
    ref: Optional[Dict[str, pd.Series]],
    x_name: str = "x",
    xaxis_name: str = "",
    yaxis_name: str = "",
    color_options: ColorOptions,
):
    cols = 1
    subplot_titles: Union[list, str] = ""

    if ref is not None:
        cols = 2
        subplot_titles = ["current", "reference"]

    fig = make_subplots(rows=1, cols=cols, shared_yaxes=True, subplot_titles=subplot_titles)
    for name, color in zip(
        ["Predicted", "Actual"], [color_options.get_current_data_color(), color_options.get_reference_data_color()]
    ):
        trace = go.Scatter(x=curr[x_name], y=curr[name], mode="lines", marker_color=color, name=name, legendgroup=name)
        fig.add_trace(trace, 1, 1)

        if ref is not None:
            trace = go.Scatter(
                x=ref[x_name],
                y=ref[name],
                mode="lines",
                marker_color=color,
                name=name,
                legendgroup=name,
                showlegend=False,
            )
            fig.add_trace(trace, 1, 2)

    # Add zero trace
    trace = go.Scatter(
        x=curr[x_name],
        y=[0] * curr[x_name].shape[0],
        mode="lines",
        marker_color=color_options.zero_line_color,
        showlegend=False,
    )
    fig.add_trace(trace, 1, 1)
    if ref is not None:
        trace = go.Scatter(
            x=ref[x_name],
            y=[0] * ref[x_name].shape[0],
            mode="lines",
            marker_color=color_options.zero_line_color,
            showlegend=False,
        )
        fig.add_trace(trace, 1, 2)
        fig.update_xaxes(title_text=xaxis_name, row=1, col=2)

    fig.update_xaxes(title_text=xaxis_name, row=1, col=1)
    fig.update_layout(yaxis_title=yaxis_name)
    fig.update_traces(marker_size=6)
    fig = json.loads(fig.to_json())
    return fig


def plot_line_in_time(
    *,
    curr: Dict[str, pd.Series],
    ref: Optional[Dict[str, pd.Series]],
    x_name: str,
    y_name: str,
    xaxis_name: str = "",
    yaxis_name: str = "",
    color_options: ColorOptions,
):
    cols = 1
    subplot_titles: Union[list, str] = ""

    if ref is not None:
        cols = 2
        subplot_titles = ["current", "reference"]

    fig = make_subplots(rows=1, cols=cols, shared_yaxes=True, subplot_titles=subplot_titles)
    trace = go.Scatter(
        x=curr[x_name],
        y=curr[y_name],
        mode="lines",
        marker_color=color_options.get_current_data_color(),
        name=y_name,
        legendgroup=y_name,
    )
    fig.add_trace(trace, 1, 1)
    # Add zero trace
    trace = go.Scatter(
        x=curr[x_name],
        y=[0] * curr[x_name].shape[0],
        mode="lines",
        marker_color=color_options.zero_line_color,
        showlegend=False,
    )
    fig.add_trace(trace, 1, 1)

    if ref is not None:
        trace = go.Scatter(
            x=ref[x_name],
            y=ref[y_name],
            mode="lines",
            marker_color=color_options.get_current_data_color(),
            name=y_name,
            legendgroup=y_name,
            showlegend=False,
        )
        fig.add_trace(trace, 1, 2)
        # Add zero trace
        trace = go.Scatter(
            x=ref[x_name],
            y=[0] * ref[x_name].shape[0],
            mode="lines",
            marker_color=color_options.zero_line_color,
            showlegend=False,
        )
        fig.add_trace(trace, 1, 2)
        fig.update_xaxes(title_text=xaxis_name, row=1, col=2)
    fig.update_xaxes(title_text=xaxis_name, row=1, col=1)
    fig.update_layout(yaxis_title=yaxis_name)
    fig.update_traces(marker_size=6)
    fig = json.loads(fig.to_json())
    return fig


def plot_error_bias_colored_scatter(
    curr_scatter_data: Dict[str, Dict[str, pd.Series]],
    ref_scatter_data: Optional[Dict[str, Dict[str, pd.Series]]],
    color_options: ColorOptions,
):
    cols = 1
    subplot_titles: Union[list, str] = ""

    if ref_scatter_data is not None:
        cols = 2
        subplot_titles = ["current", "reference"]

    fig = make_subplots(rows=1, cols=cols, shared_yaxes=True, subplot_titles=subplot_titles)

    for name, color in zip(
        ["Underestimation", "Overestimation", "Majority"],
        [color_options.underestimation_color, color_options.overestimation_color, color_options.majority_color],
    ):
        trace = go.Scatter(
            x=curr_scatter_data[name]["Actual value"],
            y=curr_scatter_data[name]["Predicted value"],
            mode="markers",
            name=name,
            legendgroup=name,
            marker_color=color
            # marker=dict(color=color_options.underestimation_color, showscale=False),
        )
        fig.add_trace(trace, 1, 1)
    fig.update_xaxes(title_text="Actual value", row=1, col=1)

    if ref_scatter_data is not None:
        for name, color in zip(
            ["Underestimation", "Overestimation", "Majority"],
            [color_options.underestimation_color, color_options.overestimation_color, color_options.majority_color],
        ):
            trace = go.Scatter(
                x=ref_scatter_data[name]["Actual value"],
                y=ref_scatter_data[name]["Predicted value"],
                mode="markers",
                name=name,
                legendgroup=name,
                showlegend=False,
                marker_color=color
                # marker=dict(color=color_options.underestimation_color, showscale=False),
            )
            fig.add_trace(trace, 1, 2)
        fig.update_xaxes(title_text="Actual value", row=1, col=2)

    fig.update_layout(
        yaxis_title="Predicted value",
        xaxis=dict(showticklabels=True),
        yaxis=dict(showticklabels=True),
    )
    fig = json.loads(fig.to_json())
    return fig


def plot_scatter_for_data_drift(
    curr_y: list, curr_x: list, y0: float, y1: float, y_name: str, x_name: str, color_options: ColorOptions
):
    fig = go.Figure()
    fig.add_trace(
        go.Scattergl(
            x=curr_x,
            y=curr_y,
            mode="markers",
            name="Current",
            marker=dict(size=6, color=color_options.get_current_data_color()),
        )
    )

    x0 = np.max(curr_x)

    fig.add_trace(
        go.Scattergl(
            x=[x0, x0],
            y=[y0, y1],
            mode="markers",
            name="Current",
            marker=dict(size=0.01, color=color_options.non_visible_color, opacity=0.005),
            showlegend=False,
        )
    )

    fig.update_layout(
        xaxis_title=x_name,
        yaxis_title=y_name,
        showlegend=True,
        legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1),
        shapes=[
            dict(
                type="rect",
                # x-reference is assigned to the x-values
                xref="paper",
                # y-reference is assigned to the plot paper [0,1]
                yref="y",
                x0=0,
                y0=y0,
                x1=1,
                y1=y1,
                fillcolor=color_options.fill_color,
                opacity=0.5,
                layer="below",
                line_width=0,
            ),
            dict(
                type="line",
                name="Reference",
                xref="paper",
                yref="y",
                x0=0,  # min(testset_agg_by_date.index),
                y0=(y0 + y1) / 2,
                x1=1,  # max(testset_agg_by_date.index),
                y1=(y0 + y1) / 2,
                line=dict(color=color_options.zero_line_color, width=3),
            ),
        ],
    )
    return fig


def plot_conf_mtrx(curr_mtrx, ref_mtrx):
    if ref_mtrx is not None:
        cols = 2
        subplot_titles = ["current", "reference"]
    else:
        cols = 1
        subplot_titles = [""]
    fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
    trace = go.Heatmap(
        z=curr_mtrx.values,
        x=[str(item) for item in curr_mtrx.labels],
        y=[str(item) for item in curr_mtrx.labels],
        text=np.array(curr_mtrx.values).astype(str),
        texttemplate="%{text}",
        coloraxis="coloraxis",
    )
    fig.add_trace(trace, 1, 1)

    if ref_mtrx is not None:
        trace = go.Heatmap(
            z=ref_mtrx.values,
            x=[str(item) for item in ref_mtrx.labels],
            y=[str(item) for item in ref_mtrx.labels],
            text=np.array(ref_mtrx.values).astype(str),
            texttemplate="%{text}",
            coloraxis="coloraxis",
        )
        fig.add_trace(trace, 1, 2)
    fig.update_layout(coloraxis={"colorscale": "RdBu_r"})
    return fig



import abc

from evidently.base_metric import InputData
from evidently.utils.data_operations import DatasetColumns


class TestPreset:
    def __init__(self):
        pass

    @abc.abstractmethod
    def generate_tests(self, data: InputData, columns: DatasetColumns):
        raise NotImplementedError()



from .classification_performance import ClassificationPreset
from .data_drift import DataDriftPreset
from .data_quality import DataQualityPreset
from .regression_performance import RegressionPreset
from .target_drift import TargetDriftPreset
from .text_overview import TextOverviewPreset

__all__ = [
    ClassificationPreset.__name__,
    DataDriftPreset.__name__,
    DataQualityPreset.__name__,
    RegressionPreset.__name__,
    TargetDriftPreset.__name__,
    TextOverviewPreset.__name__,
]



#!/usr/bin/env python
# coding: utf-8
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.utils.data_operations import process_columns


class ClassTargetNameWidget(Widget):
    def analyzers(self):
        return []

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        columns = process_columns(reference_data, column_mapping)

        if (
            columns.utility_columns.target is None
            or columns.utility_columns.prediction is None
        ):
            raise ValueError(
                f"Widget [{self.title}] requires 'target' and 'prediction' columns"
            )

        return BaseWidgetInfo(
            title=self.title,
            type="counter",
            size=2,
            params={
                "counters": [
                    {
                        "value": "",
                        "label": f"Classification Model Performance Report. Target:'{columns.utility_columns.target}'",
                    }
                ]
            },
        )



import dataclasses
import uuid
from typing import Dict
from typing import List
from typing import Optional

from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class BaseRenderer:
    """Base class for all renderers"""

    color_options: ColorOptions

    def __init__(self, color_options: Optional[ColorOptions] = None) -> None:
        if color_options is None:
            self.color_options = ColorOptions()

        else:
            self.color_options = color_options


@dataclasses.dataclass
class DetailsInfo:
    title: str
    info: BaseWidgetInfo
    id: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))


class MetricRenderer(BaseRenderer):
    def render_html(self, obj) -> List[BaseWidgetInfo]:
        raise NotImplementedError()

    def render_json(self, obj) -> dict:
        raise NotImplementedError()


@dataclasses.dataclass
class TestHtmlInfo:
    name: str
    description: str
    status: str
    details: List[DetailsInfo]
    groups: Dict[str, str]

    def with_details(self, title: str, info: BaseWidgetInfo):
        self.details.append(DetailsInfo(title, info))
        return self


class TestRenderer(BaseRenderer):
    def html_description(self, obj):
        return obj.get_result().description

    def json_description(self, obj):
        return obj.get_result().description

    def render_html(self, obj) -> TestHtmlInfo:
        result = obj.get_result()
        return TestHtmlInfo(
            name=result.name,
            description=self.html_description(obj),
            status=result.status,
            details=[],
            groups=result.groups,
        )

    def render_json(self, obj) -> dict:
        result = obj.get_result()
        return {
            "name": result.name,
            "description": self.json_description(obj),
            "status": result.status,
            "group": obj.group,
            "parameters": {},
        }


@dataclasses.dataclass
class RenderersDefinitions:
    typed_renderers: dict = dataclasses.field(default_factory=dict)
    default_html_test_renderer: Optional[TestRenderer] = None
    default_html_metric_renderer: Optional[MetricRenderer] = None


def default_renderer(wrap_type):
    def wrapper(cls):
        DEFAULT_RENDERERS.typed_renderers[wrap_type] = cls()
        return cls

    return wrapper


DEFAULT_RENDERERS = RenderersDefinitions(default_html_test_renderer=TestRenderer())



import dataclasses
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import pandas as pd

from evidently import ColumnMapping
from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_integration import get_number_of_all_pandas_missed_values
from evidently.calculations.data_integration import get_number_of_almost_constant_columns
from evidently.calculations.data_integration import get_number_of_almost_duplicated_columns
from evidently.calculations.data_integration import get_number_of_constant_columns
from evidently.calculations.data_integration import get_number_of_duplicated_columns
from evidently.calculations.data_integration import get_number_of_empty_columns
from evidently.calculations.data_quality import get_rows_count
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import table_data
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class DatasetSummary:
    """Columns information in a dataset"""

    target: Optional[str]
    prediction: Optional[Union[str, Sequence[str]]]
    date_column: Optional[str]
    id_column: Optional[str]
    number_of_columns: int
    number_of_rows: int
    number_of_missing_values: int
    number_of_categorical_columns: int
    number_of_numeric_columns: int
    number_of_text_columns: int
    number_of_datetime_columns: int
    number_of_constant_columns: int
    number_of_almost_constant_columns: int
    number_of_duplicated_columns: int
    number_of_almost_duplicated_columns: int
    number_of_empty_rows: int
    number_of_empty_columns: int
    number_of_duplicated_rows: int
    columns_type: dict
    nans_by_columns: dict
    number_uniques_by_columns: dict


@dataclasses.dataclass
class DatasetSummaryMetricResult:
    almost_duplicated_threshold: float
    current: DatasetSummary
    reference: Optional[DatasetSummary] = None


class DatasetSummaryMetric(Metric[DatasetSummaryMetricResult]):
    """Common dataset(s) columns/features characteristics"""

    # threshold for calculating the number of almost duplicated columns
    almost_duplicated_threshold: float
    almost_constant_threshold: float

    def __init__(self, almost_duplicated_threshold: float = 0.95, almost_constant_threshold: float = 0.95):
        self.almost_duplicated_threshold = almost_duplicated_threshold
        self.almost_constant_threshold = almost_constant_threshold

    def _calculate_dataset_common_stats(self, dataset: pd.DataFrame, column_mapping: ColumnMapping) -> DatasetSummary:
        columns = process_columns(dataset, column_mapping)
        return DatasetSummary(
            target=columns.utility_columns.target,
            prediction=columns.utility_columns.prediction,
            date_column=columns.utility_columns.date,
            id_column=columns.utility_columns.id_column,
            number_of_columns=len(dataset.columns),
            number_of_rows=get_rows_count(dataset),
            number_of_missing_values=get_number_of_all_pandas_missed_values(dataset),
            number_of_categorical_columns=len(columns.cat_feature_names),
            number_of_numeric_columns=len(columns.num_feature_names),
            number_of_text_columns=len(columns.text_feature_names),
            number_of_datetime_columns=len(columns.datetime_feature_names),
            number_of_empty_columns=get_number_of_empty_columns(dataset),
            number_of_constant_columns=get_number_of_constant_columns(dataset),
            number_of_almost_constant_columns=get_number_of_almost_constant_columns(
                dataset, self.almost_constant_threshold
            ),
            number_of_duplicated_columns=get_number_of_duplicated_columns(dataset),
            number_of_almost_duplicated_columns=get_number_of_almost_duplicated_columns(
                dataset, self.almost_duplicated_threshold
            ),
            number_of_empty_rows=dataset.isna().all(1).sum(),
            number_of_duplicated_rows=dataset.duplicated().sum(),
            columns_type=dict(dataset.dtypes.to_dict()),
            nans_by_columns=dataset.isna().sum().to_dict(),
            number_uniques_by_columns=dict(dataset.nunique().to_dict()),
        )

    def calculate(self, data: InputData) -> DatasetSummaryMetricResult:
        if self.almost_duplicated_threshold < 0.5 or self.almost_duplicated_threshold > 1:
            raise ValueError("Almost duplicated threshold should be in range [0.5, 1]")

        if self.almost_constant_threshold < 0.5 or self.almost_duplicated_threshold > 1:
            raise ValueError("Almost constant threshold should be in range [0.5, 1]")

        current = self._calculate_dataset_common_stats(data.current_data, data.column_mapping)
        reference = None

        if data.reference_data is not None:
            reference = self._calculate_dataset_common_stats(data.reference_data, data.column_mapping)

        return DatasetSummaryMetricResult(
            current=current,
            reference=reference,
            almost_duplicated_threshold=self.almost_duplicated_threshold,
        )


@default_renderer(wrap_type=DatasetSummaryMetric)
class DatasetSummaryMetricRenderer(MetricRenderer):
    def render_json(self, obj: DatasetSummaryMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        if "reference" in result and result["reference"]:
            result["reference"].pop("columns_type", None)

        if "current" in result and result["current"]:
            result["current"].pop("columns_type", None)

        return result

    @staticmethod
    def _get_table(metric_result: DatasetSummaryMetricResult) -> BaseWidgetInfo:
        column_names = ["Metric", "Current"]
        rows = (
            ["id column", metric_result.current.id_column],
            ["target column", metric_result.current.target],
            ["prediction column", metric_result.current.prediction],
            ["date column", metric_result.current.date_column],
            ["number of columns", metric_result.current.number_of_columns],
            ["number of rows", metric_result.current.number_of_rows],
            ["missing values", metric_result.current.number_of_missing_values],
            ["categorical columns", metric_result.current.number_of_categorical_columns],
            ["numeric columns", metric_result.current.number_of_numeric_columns],
            ["text columns", metric_result.current.number_of_text_columns],
            ["datetime columns", metric_result.current.number_of_datetime_columns],
            ["empty columns", metric_result.current.number_of_empty_columns],
            ["constant columns", metric_result.current.number_of_constant_columns],
            ["almost constant features", metric_result.current.number_of_almost_constant_columns],
            ["duplicated columns", metric_result.current.number_of_duplicated_columns],
            ["almost duplicated features", metric_result.current.number_of_almost_duplicated_columns],
        )
        if metric_result.reference is not None:
            column_names.append("Reference")
            rows[0].append(metric_result.reference.id_column)
            rows[1].append(metric_result.reference.target)
            rows[2].append(metric_result.reference.prediction)
            rows[3].append(metric_result.reference.date_column)
            rows[4].append(metric_result.reference.number_of_columns)
            rows[5].append(metric_result.reference.number_of_rows)
            rows[6].append(metric_result.reference.number_of_missing_values)
            rows[7].append(metric_result.reference.number_of_categorical_columns)
            rows[8].append(metric_result.reference.number_of_numeric_columns)
            rows[9].append(metric_result.reference.number_of_text_columns)
            rows[10].append(metric_result.reference.number_of_datetime_columns)
            rows[11].append(metric_result.reference.number_of_empty_columns)
            rows[12].append(metric_result.reference.number_of_constant_columns)
            rows[13].append(metric_result.reference.number_of_almost_constant_columns)
            rows[14].append(metric_result.reference.number_of_duplicated_columns)
            rows[15].append(metric_result.reference.number_of_almost_duplicated_columns)

        return table_data(title="", column_names=column_names, data=rows)

    def render_html(self, obj: DatasetSummaryMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        return [
            header_text(label="Dataset Summary"),
            self._get_table(metric_result),
        ]



import dataclasses
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
from sklearn.metrics import classification_report
from sklearn.metrics import log_loss
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score

from evidently.base_metric import InputData
from evidently.calculations.classification_performance import DatasetClassificationQuality
from evidently.calculations.classification_performance import PredictionData
from evidently.calculations.classification_performance import calculate_matrix
from evidently.calculations.classification_performance import calculate_metrics
from evidently.calculations.classification_performance import k_probability_threshold
from evidently.metrics.classification_performance.base_classification_metric import ThresholdClassificationMetric
from evidently.metrics.classification_performance.classification_quality_metric import ClassificationQualityMetric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import table_data
from evidently.utils.data_operations import process_columns


@dataclasses.dataclass
class ClassificationDummyMetricResults:
    dummy: DatasetClassificationQuality
    by_reference_dummy: Optional[DatasetClassificationQuality]
    model_quality: Optional[DatasetClassificationQuality]
    metrics_matrix: dict


class ClassificationDummyMetric(ThresholdClassificationMetric[ClassificationDummyMetricResults]):
    quality_metric: ClassificationQualityMetric

    def __init__(
        self,
        probas_threshold: Optional[float] = None,
        k: Optional[Union[float, int]] = None,
    ):
        super().__init__(probas_threshold, k)
        self.probas_threshold = probas_threshold
        self.k = k
        self.quality_metric = ClassificationQualityMetric()

    def calculate(self, data: InputData) -> ClassificationDummyMetricResults:
        quality_metric: Optional[ClassificationQualityMetric]
        dataset_columns = process_columns(data.current_data, data.column_mapping)
        target_name = dataset_columns.utility_columns.target
        prediction_name = dataset_columns.utility_columns.prediction

        if target_name is None:
            raise ValueError("The column 'target' should present")
        if prediction_name is None:
            quality_metric = None
        else:
            quality_metric = self.quality_metric

        #  dummy by current
        labels_ratio = data.current_data[target_name].value_counts(normalize=True)
        np.random.seed(0)
        dummy_preds = np.random.choice(labels_ratio.index, data.current_data.shape[0], p=labels_ratio)
        dummy_preds = pd.Series(dummy_preds)
        prediction: Optional[PredictionData] = None

        if prediction_name is not None:
            target, prediction = self.get_target_prediction_data(data.current_data, data.column_mapping)
            labels = prediction.labels

        else:
            target = data.current_data[target_name]
            labels = list(target.unique())

        current_matrix = calculate_matrix(
            target,
            dummy_preds,
            labels,
        )
        current_dummy = calculate_metrics(
            data.column_mapping,
            current_matrix,
            target,
            PredictionData(predictions=dummy_preds, prediction_probas=None, labels=labels),
        )

        metrics_matrix = classification_report(
            target,
            dummy_preds,
            output_dict=True,
        )
        threshold = 0.5

        if prediction is not None and prediction.prediction_probas is not None and len(labels) == 2:
            if self.probas_threshold is not None or self.k is not None:
                if self.probas_threshold is not None:
                    threshold = self.probas_threshold
                if self.k is not None:
                    threshold = k_probability_threshold(prediction.prediction_probas, self.k)

            current_dummy = self.correction_for_threshold(
                current_dummy, threshold, target, labels, prediction.prediction_probas.shape
            )

            # metrix matrix
            # neg label data
            if threshold == 1.0:
                coeff_recall = 1.0

            else:
                coeff_recall = min(1.0, 0.5 / (1 - threshold))

            coeff_precision = min(1.0, (1 - threshold) / 0.5)
            neg_label_precision = precision_score(target, dummy_preds, pos_label=labels[1]) * coeff_precision
            neg_label_recall = recall_score(target, dummy_preds, pos_label=labels[1]) * coeff_recall
            metrics_matrix = {
                str(labels[0]): {
                    "precision": current_dummy.precision,
                    "recall": current_dummy.recall,
                    "f1-score": current_dummy.f1,
                },
                str(labels[1]): {
                    "precision": neg_label_precision,
                    "recall": neg_label_recall,
                    "f1-score": 2 * neg_label_precision * neg_label_recall / (neg_label_precision + neg_label_recall),
                },
            }
        if prediction is not None and prediction.prediction_probas is not None:
            # dummy log_loss and roc_auc
            binaraized_target = (target.astype(str).values.reshape(-1, 1) == list(labels)).astype(int)
            dummy_prediction = np.full(prediction.prediction_probas.shape, 1 / prediction.prediction_probas.shape[1])
            current_dummy.log_loss = log_loss(binaraized_target, dummy_prediction)
            current_dummy.roc_auc = 0.5

        # dummy by reference
        by_reference_dummy: Optional[DatasetClassificationQuality] = None

        if data.reference_data is not None:
            labels_ratio = data.reference_data[target_name].value_counts(normalize=True)
            np.random.seed(1)
            dummy_preds = np.random.choice(labels_ratio.index, data.current_data.shape[0], p=labels_ratio)
            dummy_preds = pd.Series(dummy_preds)

            if prediction_name is not None:
                target, prediction = self.get_target_prediction_data(data.current_data, data.column_mapping)
                labels = prediction.labels
            else:
                target = data.current_data[target_name]
                labels = list(target.unique())

            current_matrix = calculate_matrix(
                target,
                dummy_preds,
                labels,
            )
            by_reference_dummy = calculate_metrics(
                data.column_mapping,
                current_matrix,
                target,
                PredictionData(predictions=dummy_preds, prediction_probas=None, labels=labels),
            )
            if prediction is not None and prediction.prediction_probas is not None and len(labels) == 2:
                by_reference_dummy = self.correction_for_threshold(
                    by_reference_dummy, threshold, target, labels, prediction.prediction_probas.shape
                )
            if prediction is not None and prediction.prediction_probas is not None:
                # dummy log_loss and roc_auc
                binaraized_target = (target.astype(str).values.reshape(-1, 1) == list(labels)).astype(int)
                dummy_prediction = np.full(
                    prediction.prediction_probas.shape, 1 / prediction.prediction_probas.shape[1]
                )
                if by_reference_dummy is not None:
                    by_reference_dummy.log_loss = log_loss(binaraized_target, dummy_prediction)
                    by_reference_dummy.roc_auc = 0.5

        # model quality
        model_quality: Optional[DatasetClassificationQuality] = None
        if quality_metric is not None:
            model_quality = quality_metric.get_result().current

        return ClassificationDummyMetricResults(
            dummy=current_dummy,
            by_reference_dummy=by_reference_dummy,
            model_quality=model_quality,
            metrics_matrix=metrics_matrix,
        )

    def correction_for_threshold(
        self,
        dummy_results: DatasetClassificationQuality,
        threshold: float,
        target: pd.Series,
        labels: list,
        probas_shape: tuple,
    ):
        if threshold == 1.0:
            coeff_precision = 1.0
        else:
            coeff_precision = min(1.0, 0.5 / (1 - threshold))
        coeff_recall = min(1.0, (1 - threshold) / 0.5)

        tpr: Optional[float] = None
        tnr: Optional[float] = None
        fpr: Optional[float] = None
        fnr: Optional[float] = None
        if (
            dummy_results.tpr is not None
            and dummy_results.tnr is not None
            and dummy_results.fpr is not None
            and dummy_results.fnr is not None
        ):
            tpr = dummy_results.tpr * coeff_recall
            tnr = dummy_results.tnr * coeff_precision
            fpr = dummy_results.fpr * coeff_recall
            fnr = dummy_results.fnr * coeff_precision

        return DatasetClassificationQuality(
            accuracy=dummy_results.accuracy,
            precision=dummy_results.precision * coeff_precision,
            recall=dummy_results.recall * coeff_recall,
            f1=2
            * dummy_results.precision
            * coeff_precision
            * dummy_results.recall
            * coeff_recall
            / (dummy_results.precision * coeff_precision + dummy_results.recall * coeff_recall),
            roc_auc=0.5,
            log_loss=None,
            tpr=tpr,
            tnr=tnr,
            fpr=fpr,
            fnr=fnr,
        )


@default_renderer(wrap_type=ClassificationDummyMetric)
class ClassificationDummyMetricRenderer(MetricRenderer):
    def render_json(self, obj: ClassificationDummyMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        return result

    def render_html(self, obj: ClassificationDummyMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        in_table_data = pd.DataFrame(data=["accuracy", "precision", "recall", "f1"])
        columns = ["Metric"]
        if metric_result.by_reference_dummy is not None:
            in_table_data["by_ref"] = [
                metric_result.by_reference_dummy.accuracy,
                metric_result.by_reference_dummy.precision,
                metric_result.by_reference_dummy.recall,
                metric_result.by_reference_dummy.f1,
            ]
            columns.append("Dummy (by rerefence)")

        in_table_data["by_curr"] = [
            metric_result.dummy.accuracy,
            metric_result.dummy.precision,
            metric_result.dummy.recall,
            metric_result.dummy.f1,
        ]
        if "Dummy (by rerefence)" in columns:
            columns.append("Dummy (by current)")
        else:
            columns.append("Dummy")

        if metric_result.model_quality is not None:
            in_table_data["model_quality"] = [
                metric_result.model_quality.accuracy,
                metric_result.model_quality.precision,
                metric_result.model_quality.recall,
                metric_result.model_quality.f1,
            ]
            columns.append("Model")

        return [
            header_text(label="Dummy Classification Quality"),
            table_data(column_names=columns, data=np.around(in_table_data, 3).values, title=""),
        ]



#!/usr/bin/env python
# coding: utf-8
from evidently.dashboard.tabs.base_tab import Tab
from evidently.dashboard.tabs.base_tab import Verbose
from evidently.dashboard.widgets.num_output_corr_widget import NumOutputCorrWidget
from evidently.dashboard.widgets.num_output_drift_widget import NumOutputDriftWidget
from evidently.dashboard.widgets.num_output_values_widget import NumOutputValuesWidget
from evidently.dashboard.widgets.num_target_pred_feature_table_widget import (
    NumTargetPredFeatureTable,
)


class NumTargetDriftTab(Tab):
    widgets = [
        (NumOutputDriftWidget("Target Drift"), Verbose.ALWAYS),
        (NumOutputCorrWidget("Target Correlations"), Verbose.FULL),
        (NumOutputValuesWidget("Target Values"), Verbose.FULL),
        (NumOutputDriftWidget("Prediction Drift", "prediction"), Verbose.ALWAYS),
        (NumOutputCorrWidget("Prediction Correlations", "prediction"), Verbose.FULL),
        (NumOutputValuesWidget("Prediction Values", "prediction"), Verbose.FULL),
        (
            NumTargetPredFeatureTable("Target (Prediction) Behavior By Feature"),
            Verbose.FULL,
        ),
    ]



import dataclasses
from abc import ABC
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from evidently.calculations.stattests import PossibleStatTestType
from evidently.metrics import ColumnDriftMetric
from evidently.metrics import DataDriftTable
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import DetailsInfo
from evidently.renderers.base_renderer import TestHtmlInfo
from evidently.renderers.base_renderer import TestRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.html_widgets import table_data
from evidently.renderers.render_utils import get_distribution_plot_figure
from evidently.tests.base_test import BaseCheckValueTest
from evidently.tests.base_test import GroupData
from evidently.tests.base_test import GroupingTypes
from evidently.tests.base_test import Test
from evidently.tests.base_test import TestResult
from evidently.tests.base_test import TestValueCondition
from evidently.utils.data_drift_utils import resolve_stattest_threshold
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.generators import BaseGenerator
from evidently.utils.types import Numeric

DATA_DRIFT_GROUP = GroupData("data_drift", "Data Drift", "")
GroupingTypes.TestGroup.add_value(DATA_DRIFT_GROUP)


@dataclasses.dataclass
class TestDataDriftResult(TestResult):
    features: Dict[str, Tuple[str, float, float]] = dataclasses.field(default_factory=dict)


class BaseDataDriftMetricsTest(BaseCheckValueTest, ABC):
    group = DATA_DRIFT_GROUP.id
    metric: DataDriftTable

    def __init__(
        self,
        columns: Optional[List[str]] = None,
        eq: Optional[Numeric] = None,
        gt: Optional[Numeric] = None,
        gte: Optional[Numeric] = None,
        is_in: Optional[List[Union[Numeric, str, bool]]] = None,
        lt: Optional[Numeric] = None,
        lte: Optional[Numeric] = None,
        not_eq: Optional[Numeric] = None,
        not_in: Optional[List[Union[Numeric, str, bool]]] = None,
        stattest: Optional[PossibleStatTestType] = None,
        cat_stattest: Optional[PossibleStatTestType] = None,
        num_stattest: Optional[PossibleStatTestType] = None,
        text_stattest: Optional[PossibleStatTestType] = None,
        per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
        stattest_threshold: Optional[float] = None,
        cat_stattest_threshold: Optional[float] = None,
        num_stattest_threshold: Optional[float] = None,
        text_stattest_threshold: Optional[float] = None,
        per_column_stattest_threshold: Optional[Dict[str, float]] = None,
    ):
        super().__init__(eq=eq, gt=gt, gte=gte, is_in=is_in, lt=lt, lte=lte, not_eq=not_eq, not_in=not_in)
        self.metric = DataDriftTable(
            columns=columns,
            stattest=stattest,
            cat_stattest=cat_stattest,
            num_stattest=num_stattest,
            text_stattest=text_stattest,
            per_column_stattest=per_column_stattest,
            stattest_threshold=stattest_threshold,
            cat_stattest_threshold=cat_stattest_threshold,
            num_stattest_threshold=num_stattest_threshold,
            text_stattest_threshold=text_stattest_threshold,
            per_column_stattest_threshold=per_column_stattest_threshold,
        )

    def check(self):
        result = super().check()
        metrics = self.metric.get_result()

        return TestDataDriftResult(
            name=result.name,
            description=result.description,
            status=result.status,
            features={
                feature: (
                    data.stattest_name,
                    np.round(data.drift_score, 3),
                    data.threshold,
                    "Detected" if data.drift_detected else "Not Detected",
                )
                for feature, data in metrics.drift_by_columns.items()
            },
        )


class TestNumberOfDriftedColumns(BaseDataDriftMetricsTest):
    name = "Number of Drifted Features"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        else:
            return TestValueCondition(lt=max(0, self.metric.get_result().number_of_columns // 3))

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().number_of_drifted_columns

    def get_description(self, value: Numeric) -> str:
        n_features = self.metric.get_result().number_of_columns
        return (
            f"The drift is detected for {value} out of {n_features} features. "
            f"The test threshold is {self.get_condition()}."
        )


class TestShareOfDriftedColumns(BaseDataDriftMetricsTest):
    name = "Share of Drifted Columns"

    def get_condition(self) -> TestValueCondition:
        if self.condition.has_condition():
            return self.condition
        else:
            return TestValueCondition(lt=0.3)

    def calculate_value_for_test(self) -> Numeric:
        return self.metric.get_result().share_of_drifted_columns

    def get_description(self, value: Numeric) -> str:
        n_drifted_features = self.metric.get_result().number_of_drifted_columns
        n_features = self.metric.get_result().number_of_columns
        return (
            f"The drift is detected for {value * 100:.3g}% features "
            f"({n_drifted_features} out of {n_features}). The test threshold is {self.get_condition()}"
        )


class TestColumnDrift(Test):
    name = "Drift per Column"
    group = DATA_DRIFT_GROUP.id
    metric: ColumnDriftMetric
    column_name: str

    def __init__(
        self,
        column_name: str,
        stattest: Optional[PossibleStatTestType] = None,
        stattest_threshold: Optional[float] = None,
    ):
        self.column_name = column_name
        self.metric = ColumnDriftMetric(
            column_name=column_name, stattest=stattest, stattest_threshold=stattest_threshold
        )

    def check(self):
        drift_info = self.metric.get_result()

        p_value = np.round(drift_info.drift_score, 3)
        stattest_name = drift_info.stattest_name
        threshold = drift_info.stattest_threshold
        description = (
            f"The drift score for the feature **{self.column_name}** is {p_value:.3g}. "
            f"The drift detection method is {stattest_name}. "
            f"The drift detection threshold is {threshold}."
        )

        if not drift_info.drift_detected:
            result_status = TestResult.SUCCESS

        else:
            result_status = TestResult.FAIL

        return TestResult(
            name=self.name,
            description=description,
            status=result_status,
            groups={
                GroupingTypes.ByFeature.id: self.column_name,
            },
        )


class TestAllFeaturesValueDrift(BaseGenerator):
    """Create value drift tests for numeric and category features"""

    columns: Optional[List[str]]
    stattest: Optional[PossibleStatTestType]
    cat_stattest: Optional[PossibleStatTestType]
    num_stattest: Optional[PossibleStatTestType]
    text_stattest: Optional[PossibleStatTestType]
    per_column_stattest: Optional[Dict[str, PossibleStatTestType]]
    stattest_threshold: Optional[float]
    cat_stattest_threshold: Optional[float]
    num_stattest_threshold: Optional[float]
    text_stattest_threshold: Optional[float]
    per_column_stattest_threshold: Optional[Dict[str, float]]

    def __init__(
        self,
        columns: Optional[List[str]] = None,
        stattest: Optional[PossibleStatTestType] = None,
        cat_stattest: Optional[PossibleStatTestType] = None,
        num_stattest: Optional[PossibleStatTestType] = None,
        text_stattest: Optional[PossibleStatTestType] = None,
        per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
        stattest_threshold: Optional[float] = None,
        cat_stattest_threshold: Optional[float] = None,
        num_stattest_threshold: Optional[float] = None,
        text_stattest_threshold: Optional[float] = None,
        per_column_stattest_threshold: Optional[Dict[str, float]] = None,
    ):
        self.columns = columns
        self.stattest = stattest
        self.cat_stattest = cat_stattest
        self.num_stattest = num_stattest
        self.text_stattest = text_stattest
        self.per_column_stattest = per_column_stattest
        self.stattest_threshold = stattest_threshold
        self.cat_stattest_threshold = cat_stattest_threshold
        self.num_stattest_threshold = num_stattest_threshold
        self.text_stattest_threshold = text_stattest_threshold
        self.per_column_stattest_threshold = per_column_stattest_threshold

    def generate(self, columns_info: DatasetColumns) -> List[TestColumnDrift]:
        results = []
        for name in columns_info.cat_feature_names:
            if self.columns and name not in self.columns:
                continue
            stattest, threshold = resolve_stattest_threshold(
                name,
                "cat",
                self.stattest,
                self.cat_stattest,
                self.num_stattest,
                self.text_stattest,
                self.per_column_stattest,
                self.stattest_threshold,
                self.cat_stattest_threshold,
                self.num_stattest_threshold,
                self.text_stattest_threshold,
                self.per_column_stattest_threshold,
            )
            results.append(TestColumnDrift(column_name=name, stattest=stattest, stattest_threshold=threshold))
        for name in columns_info.num_feature_names:
            if self.columns and name not in self.columns:
                continue
            stattest, threshold = resolve_stattest_threshold(
                name,
                "num",
                self.stattest,
                self.cat_stattest,
                self.num_stattest,
                self.text_stattest,
                self.per_column_stattest,
                self.stattest_threshold,
                self.cat_stattest_threshold,
                self.num_stattest_threshold,
                self.text_stattest_threshold,
                self.per_column_stattest_threshold,
            )
            results.append(TestColumnDrift(column_name=name, stattest=stattest, stattest_threshold=threshold))
        for name in columns_info.text_feature_names:
            if self.columns and name not in self.columns:
                continue
            stattest, threshold = resolve_stattest_threshold(
                name,
                "text",
                self.stattest,
                self.cat_stattest,
                self.num_stattest,
                self.text_stattest,
                self.per_column_stattest,
                self.stattest_threshold,
                self.cat_stattest_threshold,
                self.num_stattest_threshold,
                self.text_stattest_threshold,
                self.per_column_stattest_threshold,
            )
            results.append(TestColumnDrift(column_name=name, stattest=stattest, stattest_threshold=threshold))
        return results


class TestCustomFeaturesValueDrift(BaseGenerator):
    """Create value drift tests for specified features"""

    features: List[str]
    stattest: Optional[PossibleStatTestType] = None
    cat_stattest: Optional[PossibleStatTestType] = None
    num_stattest: Optional[PossibleStatTestType] = None
    text_stattest: Optional[PossibleStatTestType] = None
    per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None
    stattest_threshold: Optional[float] = None
    cat_stattest_threshold: Optional[float] = None
    num_stattest_threshold: Optional[float] = None
    text_stattest_threshold: Optional[float] = None
    per_column_stattest_threshold: Optional[Dict[str, float]] = None

    def __init__(
        self,
        features: List[str],
        stattest: Optional[PossibleStatTestType] = None,
        cat_stattest: Optional[PossibleStatTestType] = None,
        num_stattest: Optional[PossibleStatTestType] = None,
        text_stattest: Optional[PossibleStatTestType] = None,
        per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
        stattest_threshold: Optional[float] = None,
        cat_stattest_threshold: Optional[float] = None,
        num_stattest_threshold: Optional[float] = None,
        text_stattest_threshold: Optional[float] = None,
        per_column_stattest_threshold: Optional[Dict[str, float]] = None,
    ):
        self.features = features
        self.stattest = stattest
        self.cat_stattest = cat_stattest
        self.num_stattest = num_stattest
        self.text_stattest = text_stattest
        self.per_column_stattest = per_column_stattest
        self.stattest_threshold = stattest_threshold
        self.cat_stattest_threshold = cat_stattest_threshold
        self.num_stattest_threshold = num_stattest_threshold
        self.text_stattest_threshold = text_stattest_threshold
        self.per_feature_threshold = per_column_stattest_threshold

    def generate(self, columns_info: DatasetColumns) -> List[TestColumnDrift]:
        result = []
        for name in self.features:
            stattest, threshold = resolve_stattest_threshold(
                name,
                "cat"
                if name in columns_info.cat_feature_names
                else "num"
                if columns_info.num_feature_names
                else "text"
                if columns_info.text_feature_names
                else "datetime",
                self.stattest,
                self.cat_stattest,
                self.num_stattest,
                self.text_stattest,
                self.per_column_stattest,
                self.stattest_threshold,
                self.cat_stattest_threshold,
                self.num_stattest_threshold,
                self.text_stattest_threshold,
                self.per_column_stattest_threshold,
            )
            result.append(
                TestColumnDrift(
                    column_name=name,
                    stattest=stattest,
                    stattest_threshold=threshold,
                )
            )
        return result


@default_renderer(wrap_type=TestNumberOfDriftedColumns)
class TestNumberOfDriftedColumnsRenderer(TestRenderer):
    def render_json(self, obj: TestNumberOfDriftedColumns) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["features"] = {
            feature: {"stattest": data[0], "score": np.round(data[1], 3), "threshold": data[2], "data_drift": data[3]}
            for feature, data in obj.get_result().features.items()
        }
        return base

    def render_html(self, obj: TestNumberOfDriftedColumns) -> TestHtmlInfo:
        info = super().render_html(obj)
        df = pd.DataFrame(
            data=[[feature] + list(data) for feature, data in obj.get_result().features.items()],
            columns=["Feature name", "Stattest", "Drift score", "Threshold", "Data Drift"],
        )
        df = df.sort_values("Data Drift")
        info.with_details(title="Drift Table", info=table_data(column_names=df.columns.to_list(), data=df.values))
        return info


@default_renderer(wrap_type=TestShareOfDriftedColumns)
class TestShareOfDriftedColumnsRenderer(TestRenderer):
    def render_json(self, obj: TestShareOfDriftedColumns) -> dict:
        base = super().render_json(obj)
        base["parameters"]["condition"] = obj.get_condition().as_dict()
        base["parameters"]["features"] = {
            feature: {"stattest": data[0], "score": np.round(data[1], 3), "threshold": data[2], "data_drift": data[3]}
            for feature, data in obj.get_result().features.items()
        }
        return base

    def render_html(self, obj: TestShareOfDriftedColumns) -> TestHtmlInfo:
        info = super().render_html(obj)
        df = pd.DataFrame(
            data=[[feature] + list(data) for feature, data in obj.get_result().features.items()],
            columns=["Feature name", "Stattest", "Drift score", "Threshold", "Data Drift"],
        )
        df = df.sort_values("Data Drift")
        info.details = [
            DetailsInfo(
                id="drift_table",
                title="",
                info=BaseWidgetInfo(
                    title="",
                    type="table",
                    params={"header": df.columns.to_list(), "data": df.values},
                    size=2,
                ),
            ),
        ]
        return info


@default_renderer(wrap_type=TestColumnDrift)
class TestColumnDriftRenderer(TestRenderer):
    def render_json(self, obj: TestColumnDrift) -> dict:
        feature_name = obj.column_name
        drift_data = obj.metric.get_result()
        base = super().render_json(obj)
        base["parameters"]["features"] = {
            feature_name: {
                "stattest_name": drift_data.stattest_name,
                "score": np.round(drift_data.drift_score, 3),
                "stattest_threshold": drift_data.stattest_threshold,
                "data_drift": drift_data.drift_detected,
            }
        }
        return base

    def render_html(self, obj: TestColumnDrift) -> TestHtmlInfo:
        info = super().render_html(obj)
        result = obj.metric.get_result()
        column_name = obj.column_name
        if result.column_type == "text":
            if result.typical_words_cur is not None and result.typical_words_ref is not None:
                info.details = [
                    DetailsInfo(
                        id=f"{column_name} dritf curr",
                        title="current: characteristic words",
                        info=BaseWidgetInfo(
                            title="",
                            type="table",
                            params={"header": ["", ""], "data": [[el, ""] for el in result.typical_words_cur]},
                            size=2,
                        ),
                    ),
                    DetailsInfo(
                        id=f"{column_name} dritf ref",
                        title="reference: characteristic words",
                        info=BaseWidgetInfo(
                            title="",
                            type="table",
                            params={"header": ["", ""], "data": [[el, ""] for el in result.typical_words_ref]},
                            size=2,
                        ),
                    ),
                ]
            else:
                return info
        else:
            if result.current_distribution is None:
                raise ValueError("Expected data is missing")
            fig = get_distribution_plot_figure(
                current_distribution=result.current_distribution,
                reference_distribution=result.reference_distribution,
                color_options=self.color_options,
            )
            info.with_details(f"{column_name}", plotly_figure(title="", figure=fig))
        return info
